# Generated by Django 4.0.8 on 2022-12-07 14:46

from decimal import Decimal

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0043_priceconfig_eve_market_group_alter_priceconfig_price'),
    ]

    operations = [
        migrations.AddField(
            model_name='priceconfig',
            name='minimum',
            field=models.DecimalField(decimal_places=2, default=Decimal('0'), help_text='Minimum isk to consider an order at', max_digits=15, verbose_name='Price'),
        ),
    ]



from django.utils.html import format_html


def list_2_html_w_tooltips(my_items: list, max_items: int) -> str:
    """converts list of strings into HTML with cutoff and tooltip"""
    items_truncated_str = ', '.join(my_items[:max_items])
    if not my_items:
        result = None
    elif len(my_items) <= max_items:
        result = items_truncated_str
    else:
        items_truncated_str += ', (...)'
        items_all_str = ', '.join(my_items)
        result = format_html(
            '<span data-tooltip="{}" class="tooltip">{}</span>',
            items_all_str,
            items_truncated_str
        )
    return result



# Generated by Django 4.0.3 on 2022-03-21 14:06

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0027_managedwatchconfig_remove_watchconfig_managed_app_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='managedwatchconfig',
            old_name='manged_app_reason',
            new_name='managed_app_reason',
        ),
    ]



# Generated by Django 4.0.6 on 2022-08-01 11:49

from decimal import Decimal

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0007_evetype_description'),
        ('marketmanager', '0028_rename_manged_app_reason_managedwatchconfig_managed_app_reason'),
    ]

    operations = [
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_buy_order',
            field=models.BooleanField(default=False, verbose_name='Buy Order'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_debug_webhooks',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on WatchConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='managed_debug_webhooks', to='marketmanager.webhook', verbose_name='Debug Webhook'),
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_jita_compare_percent',
            field=models.DecimalField(decimal_places=0, default=Decimal('0'), help_text='If set ignores Flat price value', max_digits=3, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)], verbose_name='Jita Comparison %'),
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_quantity',
            field=models.IntegerField(default=1, help_text='Apply a multiplier to the volumes that will be generated', verbose_name='quantity'),
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_region',
            field=models.ManyToManyField(blank=True, to='eveuniverse.everegion', verbose_name='Region'),
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_solar_system',
            field=models.ManyToManyField(blank=True, to='eveuniverse.evesolarsystem', verbose_name='Solar System'),
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_structure',
            field=models.ManyToManyField(blank=True, to='marketmanager.structure', verbose_name='Structure'),
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_structure_type',
            field=models.ManyToManyField(blank=True, help_text='Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)', related_name='managed_structure_type', to='eveuniverse.evetype', verbose_name='Structure Type Filter'),
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_webhooks',
            field=models.ManyToManyField(blank=True, to='marketmanager.webhook', verbose_name='Webhooks'),
        ),
    ]



def _dt_eveformat(dt: object) -> str:
    """converts a datetime to a string in eve format
    e.g. '2019-06-25T19:04:44'
    """
    from datetime import datetime

    dt2 = datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)
    return dt2.isoformat()


def _generate_token(
    character_id: int,
    character_name: str,
    access_token: str = 'access_token',
    refresh_token: str = 'refresh_token',
    character_owner_hash: str = 'character_owner_hash',
    scopes=None,  # list or string from ccp
    timestamp_dt: object = None,
    expires_in: int = 1200,
    sso_version: int = 2
) -> dict:
    from datetime import datetime, timedelta

    if timestamp_dt is None:
        timestamp_dt = datetime.utcnow()
    if scopes is None:
        scopes = [
            'esi-mail.read_mail.v1',
            'esi-wallet.read_character_wallet.v1',
            'esi-universe.read_structures.v1'
        ]
    token = {
        'access_token': access_token,
        'expires_in': expires_in,
        'refresh_token': refresh_token,
        'timestamp': int(timestamp_dt.timestamp()),
        'character_id': character_id,
        'name': character_name,
        'ExpiresOn': _dt_eveformat(timestamp_dt + timedelta(seconds=expires_in)),
        'scp': scopes,
        'token_type': 'character',
        'owner': character_owner_hash,
        'IntellectualProperty': 'EVE',
        'sso_version': sso_version
    }
    return token


def _store_as_Token(token: dict, user: object) -> object:
    """Stores a generated token dict as Token object for given user

    returns Token object
    """
    from esi.models import Scope, Token

    obj = Token.objects.create(
        access_token=token['access_token'],
        refresh_token=token['refresh_token'],
        user=user,
        character_id=token['character_id'],
        character_name=token['name'],
        token_type=token['token_type'],
        character_owner_hash=token['owner'],
        sso_version=token['sso_version']
    )

    if isinstance(token['scp'], str):
        token['scp'] = [token['scp']]

    for scope_name in token['scp']:
        scope, _ = Scope.objects.get_or_create(
            name=scope_name
        )
        obj.scopes.add(scope)

    return obj



# Generated by Django 4.0.6 on 2022-08-06 06:35

from decimal import Decimal

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0032_alter_watchconfig_managed_watch_config'),
    ]

    operations = [
        migrations.AlterField(
            model_name='managedwatchconfig',
            name='managed_jita_compare_percent',
            field=models.DecimalField(decimal_places=0, default=Decimal('0'), help_text='If set ignores Flat price value', max_digits=4, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(1000)], verbose_name='Jita Comparison %'),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-14 12:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0005_alter_watchconfig_constellation_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='watchconfig',
            name='config_type',
            field=models.CharField(choices=[('Supply', 'Supply Check'), ('Bargain', 'Bargain Finder'), ('Scalp', 'Scalp Finder')], default='Supply', help_text='The type of WatchConfig, check documentation for more.', max_length=10, verbose_name='Config Type'),
        ),
        migrations.AlterField(
            model_name='watchconfig',
            name='webhooks',
            field=models.ManyToManyField(blank=True, to='marketmanager.webhook', verbose_name='Webhooks'),
        ),
    ]



# Generated by Django 4.0.6 on 2022-08-06 07:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0035_alter_watchconfig_jita_compare_percent'),
    ]

    operations = [
        migrations.AddField(
            model_name='watchconfig',
            name='last_result_volume',
            field=models.IntegerField(blank=True, help_text='The Volume returned the last time this WatchConfig was run', null=True, verbose_name='Last Result Volume'),
        ),
        migrations.AddField(
            model_name='watchconfig',
            name='last_run',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Last Task Runtime'),
        ),
    ]



# Generated by Django 3.2.8 on 2021-12-01 15:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0006_auto_20211201_1343'),
    ]

    operations = [
        migrations.AlterField(
            model_name='order',
            name='is_buy_order',
            field=models.BooleanField(db_index=True, default=False, help_text='True if the order is a bid (buy) order', verbose_name='Buy Order'),
        ),
    ]



from allianceauth import __version__ as aa__version__
from allianceauth.services.hooks import get_extension_logger
from esi.clients import EsiClientProvider
from esi.models import Token

from . import __version__ as mm__version__

logger = get_extension_logger(__name__)

APP_INFO_TEXT = f"allianceauth v{aa__version__} & aa-market-manager v{mm__version__}"

"""
Swagger spec operations:
get_universe_structures
get_universe_structures_structure_id
get_markets_region_id_orders
get_markets_structures_structure_id
get_characters_character_id_orders
get_characters_character_id_orders_history
get_corporations_corporation_id_orders
get_corporations_corporation_id_orders_history
get_corporations_corporation_id_structures
"""

esi = EsiClientProvider(app_info_text=APP_INFO_TEXT)


def get_universe_structures(filter: str = "all"):
    result = esi.client.Universe.get_universe_structures(
        filter=filter
    ).results()
    return result


def get_universe_structures_structure_id(structure_id: int, token: Token):
    return esi.client.Universe.get_universe_structures_structure_id(structure_id=structure_id, token=token.valid_access_token()).results()


def get_markets_region_id_orders(
        region_id: int,
        order_type: str = "all"):
    result = esi.client.Market.get_markets_region_id_orders(
        order_type=order_type,
        region_id=region_id
    ).results()
    return result


def get_markets_region_id_orders_paged(
        region_id: int,
        page: int,
        order_type: str = "all"):
    result = esi.client.Market.get_markets_region_id_orders(
        order_type=order_type,
        region_id=region_id,
        page=page
    )
    result.request_config.also_return_response = True
    return result.result()


def get_markets_region_id_orders_by_typeid(
        region_id: int,
        order_type: str = "all",
        type_id: str = ""):
    result = esi.client.Market.get_markets_region_id_orders(
        order_type=order_type,
        region_id=region_id,
        type_id=type_id
    ).results()
    return result


def get_markets_region_id_orders_by_typeid_paged(
        region_id: int,
        page: int,
        order_type: str = "all",
        type_id: str = ""):
    result = esi.client.Market.get_markets_region_id_orders(
        order_type=order_type,
        region_id=region_id,
        type_id=type_id,
        page=page
    )
    result.request_config.also_return_response = True
    return result.result()


def get_markets_region_id_history(
        region_id: int,
        order_type: str = "all",
        type_id: str = ""):
    result = esi.client.Market.get_markets_region_id_history(
        order_type=order_type,
        region_id=region_id,
        type_id=type_id
    ).results()
    return result


def get_markets_structures_structure_id(structure_id: int, token: Token):
    result = esi.client.Market.get_markets_structures_structure_id(
        structure_id=structure_id,
        token=token.valid_access_token()
    ).results()
    return result


def get_characters_character_id_orders(character_id: int):
    required_scopes = ['esi-markets.read_character_orders.v1']
    token = Token.get_token(character_id, required_scopes)

    result = esi.client.Market.get_characters_character_id_orders(
        character_id=character_id,
        token=token.valid_access_token()
    ).results()
    return result


def get_characters_character_id_orders_history(character_id: int):
    required_scopes = ['esi-markets.read_character_orders.v1']
    token = Token.get_token(character_id, required_scopes)

    result = esi.client.Market.get_characters_character_id_orders_history(
        character_id=character_id,
        token=token.valid_access_token()
    ).results()
    return result


def get_characters_character_id_roles_from_token(token: Token):
    # Yes this is weird, its because im pulling _specific_ scopes to find this token elsewhere
    result = esi.client.Character.get_characters_character_id_roles(
        character_id=token.character_id,
        token=token.valid_access_token()
    ).results()
    return result


def get_corporations_corporation_id_orders(corporation_id: int, token: Token):
    result = esi.client.Market.get_corporations_corporation_id_orders(
        corporation_id=corporation_id,
        token=token.valid_access_token()
    ).results()
    return result


def get_corporations_corporation_id_orders_history(corporation_id: int, token: Token):
    result = esi.client.Market.get_corporations_corporation_id_orders_history(
        character_id=corporation_id,
        token=token.valid_access_token()
    ).results()
    return result


def get_corporations_corporation_id_structures(corporation_id: int, token: Token):
    result = esi.client.Corporation.get_corporations_corporation_id_structures(
        corporation_id=corporation_id,
        token=token.valid_access_token()
    ).results()
    return result



import datetime
import json
from typing import Iterable

from eveuniverse.models import EveEntity, EveMarketGroup, EveRegion, EveType

from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F, Sum
from django.db.models.expressions import Case, When
from django.http.response import HttpResponse, JsonResponse
from django.shortcuts import redirect, render
from django.template import TemplateDoesNotExist

from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required

from marketmanager.app_settings import (
    MARKETMANAGER_TYPESTATISTICS_MINIMUM_ORDER_COUNT,
)
from marketmanager.models import (
    Order, PublicConfig, Structure, SupplyConfig, TypeStatistics,
)

logger = get_extension_logger(__name__)

CHARACTER_SCOPES = [
    'esi-markets.read_character_orders.v1',
    'esi-markets.structure_markets.v1',
    'esi-universe.read_structures.v1',
]

CORPORATION_SCOPES = [
    'esi-markets.read_corporation_orders.v1',
    'esi-markets.structure_markets.v1',
    'esi-characters.read_corporation_roles.v1',
    'esi-corporations.read_structures.v1',
    'esi-universe.read_structures.v1',
]


@login_required
@permission_required("marketmanager.basic_market_browser")
def marketbrowser(request) -> HttpResponse:
    region_id = request.GET.get('region_id', None)
    type_id = request.GET.get('type_id', None)
    all_regions = PublicConfig.get_solo().fetch_regions.all()
    parent_market_groups = EveMarketGroup.objects.filter(
        parent_market_group_id__isnull=True)

    try:
        eve_type, eve_type_fetched = EveType.objects.get_or_create_esi(
            id=type_id)
    except Exception:
        eve_type = None

    try:
        eve_region, everegion_fetched = EveRegion.objects.get_or_create_esi(
            id=region_id)
    except Exception:
        eve_region = None

    if eve_type is not None:
        eve_type_icon_url = eve_type.icon_url(size=256)
    else:
        eve_type_icon_url = None

    render_items = {
        "all_regions": all_regions,
        "parent_market_groups": parent_market_groups,
        "eve_region": eve_region,
        "eve_type": eve_type,
        "eve_type_icon_url": eve_type_icon_url,
        "type_statistics": type_statistics(eve_type=eve_type, eve_region=eve_region),
    }
    try:
        # AA 4.x
        return render(request, "marketmanager/marketbrowser-bs5.html", render_items)
    except TemplateDoesNotExist:
        # AA 3.x
        return render(request, "marketmanager/marketbrowser.html", render_items)


@login_required
@permission_required("marketmanager.basic_market_watches")
def marketwatches(request) -> HttpResponse:

    watchconfigs = SupplyConfig.objects.all().annotate(
        missing_volume=F("volume") - F("last_result_volume")
    ).filter(
        missing_volume__gt=0)
    render_items = {
        "watchconfigs": watchconfigs,
    }
    try:
        # AA 4.x
        return render(request, "marketmanager/marketwatches-bs5.html", render_items)
    except TemplateDoesNotExist:
        # AA 3.x
        return render(request, "marketmanager/marketwatches.html", render_items)


@login_required
@permission_required("marketmanager.basic_market_browser")
def marketbrowser_autocomplete(request) -> HttpResponse:
    if request.headers.get('x-requested-with') == 'XMLHttpRequest':  # is_ajax
        search_query = request.GET.get('term')

    autocomplete_query = EveType.objects.filter(
        name__icontains=search_query,
        eve_market_group__isnull=False,
        published=1
    ).order_by("name")

    result = []
    for possible in autocomplete_query:
        data = {}
        data['label'] = possible.name
        data['value'] = possible.id
        result.append(data)
    dump = json.dumps(result)

    mimetype = 'application/json'
    return HttpResponse(dump, mimetype)


@login_required
@permission_required("marketmanager.basic_market_browser")
def marketbrowser_buy_orders(request):
    if request.headers.get('x-requested-with') == 'XMLHttpRequest':  # is_ajax
        region_id = request.GET.get('region_id', None)
        type_id = request.GET.get('type_id', None)
    else:
        region_id = None
        type_id = None

    if request.user.has_perm("marketmanager.order_highlight_user"):
        user_characters = request.user.character_ownerships.all(
        ).select_related('character').values('character')
    else:
        user_characters = []

    if request.user.has_perm("marketmanager.order_highlight_corporation"):
        user_corporation_ids = request.user.character_ownerships.all().select_related(
            'character__corporation_id').values('character__corporation_id')
    else:
        user_corporation_ids = []

    buy_orders = Order.objects.filter(
        eve_type=type_id,
        is_buy_order=True
    ).annotate(
        user_is_owner=Case(
            When(
                issued_by_character__in=user_characters,
                then=True
            ),
            default=False
        ),
        corporation_is_owner=Case(
            When(
                issued_by_corporation__corporation_id__in=user_corporation_ids,
                then=True
            ),
            default=False
        ),
    ).values(
        'volume_remain',
        'price',
        'location_id',
        'issued',
        'duration',
        'eve_region__name',
        'updated_at',
        'user_is_owner',
        'corporation_is_owner'
    )
    if region_id is not None:
        buy_orders = buy_orders.filter(
            eve_region=region_id
        )

    buy_order_locations = []
    for order in buy_orders:
        buy_order_locations.append(order['location_id'])
    eveentities_resolved, structures_resolved = bulk_location_resolver(
        buy_order_locations)

    for order in buy_orders:
        if eveentities_resolved.to_name(order['location_id']) != "":
            order['location_resolved'] = eveentities_resolved.to_name(
                order['location_id'])
        else:
            try:
                order['location_resolved'] = structures_resolved[order['location_id']].name
            except KeyError:
                order['location_resolved'] = order['location_id']
        order["expiry_calculated"] = order["issued"] + \
            datetime.timedelta(days=order["duration"])

    return JsonResponse({"buy_orders": list(buy_orders)})


@login_required
@permission_required("marketmanager.basic_market_browser")
def marketbrowser_sell_orders(request):
    if request.headers.get('x-requested-with') == 'XMLHttpRequest':  # is_ajax
        region_id = request.GET.get('region_id', None)
        type_id = request.GET.get('type_id', None)
    else:
        region_id = None
        type_id = None

    if request.user.has_perm("marketmanager.order_highlight_user"):
        user_characters = request.user.character_ownerships.all(
        ).select_related('character').values('character')
    else:
        user_characters = []

    if request.user.has_perm("marketmanager.order_highlight_corporation"):
        user_corporation_ids = request.user.character_ownerships.all().select_related(
            'character__corporation_id').values('character__corporation_id')
    else:
        user_corporation_ids = []

    # Sell Orders
    sell_orders = Order.objects.filter(
        eve_type=type_id,
        is_buy_order=False
    ).annotate(
        user_is_owner=Case(
            When(
                issued_by_character__in=user_characters,
                then=True
            ),
            default=False
        ),
        corporation_is_owner=Case(
            When(
                issued_by_corporation__corporation_id__in=user_corporation_ids,
                then=True
            ),
            default=False
        ),
    ).values(
        'volume_remain',
        'price',
        'location_id',
        'issued',
        'duration',
        'eve_region__name',
        'updated_at',
        'order_id',
        'user_is_owner',
        'corporation_is_owner'
    )

    if region_id is not None:
        sell_orders = sell_orders.filter(
            eve_region=region_id
        )

    sell_order_locations = []
    for order in sell_orders:
        sell_order_locations.append(order['location_id'])
    eveentities_resolved, structures_resolved = bulk_location_resolver(
        sell_order_locations)

    for order in sell_orders:
        if eveentities_resolved.to_name(order['location_id']) != "":
            order['location_resolved'] = eveentities_resolved.to_name(
                order['location_id'])
        else:
            try:
                order['location_resolved'] = structures_resolved[order['location_id']].name
            except KeyError:
                order['location_resolved'] = order['location_id']
        order["expiry_calculated"] = order["issued"] + \
            datetime.timedelta(days=order["duration"])

    return JsonResponse({"sell_orders": list(sell_orders)})


@login_required
@permission_required("marketmanager.basic_market_browser")
def item_selector(request):
    data = EveMarketGroup.objects.all()
    return render(request, "marketmanager/item_selector.html", data)


@login_required
@token_required(scopes=CHARACTER_SCOPES)
def add_char(request, token):
    return redirect('marketmanager:marketbrowser')


@login_required
@token_required(scopes=CORPORATION_SCOPES)
def add_corp(request, token):
    return redirect('marketmanager:marketbrowser')


def location_resolver(location_id) -> str:
    if location_id >= 60000000 and location_id <= 64000000:
        # EveStation (Range: 60000000 - 64000000)
        # EveEntity has its own resolver
        # but i dont want Structures to slip through
        # and spam ESI errors
        return EveEntity.objects.resolve_name(location_id)
    else:
        try:
            return Structure.objects.get(structure_id=location_id).name
        except Exception as e:
            logger.exception(e)
            return str(location_id)


def bulk_location_resolver(location_ids: Iterable[int]):
    bulk_eve_entity_ids = []
    bulk_structure_ids = []

    for location_id in location_ids:
        if location_id >= 60000000 and location_id <= 64000000:
            # EveStation (Range: 60000000 - 64000000)
            # EveEntity has its own resolver
            # but i dont want Structures to slip through
            # and spam ESI errors
            bulk_eve_entity_ids.append(location_id)
        else:
            bulk_structure_ids.append(location_id)

    eveentity_resolver = EveEntity.objects.bulk_resolve_names(
        bulk_eve_entity_ids)
    structure_resolver = Structure.objects.in_bulk(bulk_structure_ids)

    return eveentity_resolver, structure_resolver


def type_statistics(eve_type: EveType, eve_region: EveRegion):
    # Returns a specific set of stats for the item_details template
    orders = Order.objects.filter(eve_type=eve_type)
    if eve_region is not None:
        orders = orders.filter(
            eve_region=eve_region
        )
    try:
        order_stats = TypeStatistics.objects.get(
            eve_type=eve_type, eve_region=eve_region)
    except ObjectDoesNotExist:
        return {
            'buy_fifth_percentile': 0,
            'sell_fifth_percentile': 0,
            'buy_weighted_average': 0,
            'sell_weighted_average': 0,
            'buy_median': 0,
            'sell_median': 0,
            'buy_volume': orders.filter(is_buy_order=True).aggregate(volume=Sum(F('volume_remain')))["volume"],
            'sell_volume': orders.filter(is_buy_order=False).aggregate(volume=Sum(F('volume_remain')))["volume"],
            'explain': f"TypeStatistics have not been calculated yet, or skipped due to less than {MARKETMANAGER_TYPESTATISTICS_MINIMUM_ORDER_COUNT} orders"
        }
    return {
        'buy_fifth_percentile': order_stats.buy_fifth_percentile,
        'sell_fifth_percentile': order_stats.sell_fifth_percentile,
        'buy_weighted_average': order_stats.buy_weighted_average,
        'sell_weighted_average': order_stats.sell_weighted_average,
        'buy_median': order_stats.buy_median,
        'sell_median': order_stats.sell_median,
        'buy_volume': orders.filter(is_buy_order=True).aggregate(volume=Sum(F('volume_remain')))["volume"],
        'sell_volume': orders.filter(is_buy_order=False).aggregate(volume=Sum(F('volume_remain')))["volume"],
        'explain': None
    }



# Generated by Django 4.0.3 on 2022-03-14 12:32

from decimal import Decimal

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0003_rename_type_watchconfig_eve_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='watchconfig',
            name='jita_compare_percent',
            field=models.DecimalField(decimal_places=0, default=Decimal('0'), help_text='If set ignores Flat price value', max_digits=3, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)], verbose_name='Jita Comparison %'),
        ),
        migrations.AlterField(
            model_name='watchconfig',
            name='price',
            field=models.IntegerField(default=0, help_text='Set to Zero to skip this filter', verbose_name='Price'),
        ),
        migrations.AlterField(
            model_name='watchconfig',
            name='volume',
            field=models.IntegerField(default=1, help_text='Set to Zero to check ANY/EVERY order against Price', verbose_name='Volume'),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-13 12:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0009_rename_config_publicconfig_order_source_character_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='order',
            name='source_private',
            field=models.BooleanField(default=False, verbose_name='Sourced From Public Market Data'),
        ),
    ]



# Generated by Django 4.0.8 on 2023-01-11 03:22

from decimal import Decimal

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0044_priceconfig_minimum'),
    ]

    operations = [
        migrations.AlterField(
            model_name='priceconfig',
            name='minimum',
            field=models.DecimalField(decimal_places=2, default=Decimal('1000000'), help_text='Minimum isk to consider an order at, Default: 1M', max_digits=15, verbose_name='Minimum Price'),
        ),
    ]



# Generated by Django 4.0.6 on 2022-08-06 06:37

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0033_alter_managedwatchconfig_managed_jita_compare_percent'),
    ]

    operations = [
        migrations.AlterField(
            model_name='watchconfig',
            name='managed_watch_config',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='marketmanager.managedwatchconfig', verbose_name='Managed Watch Configs'),
        ),
    ]



# Generated by Django 4.0.6 on 2022-08-06 08:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0036_watchconfig_last_result_volume_watchconfig_last_run'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='general',
            options={'default_permissions': (), 'managed': False, 'permissions': (('basic_market_browser', 'Can access the Standard Market Browser'), ('basic_market_watches', 'Can view all WatchConfigs and their status'), ('order_highlight_user', "Market orders owned by the user's characters may be highlighted in the standard/basic Market Browser"), ('order_highlight_corporation', 'Market orders owned by any corporation a user is a member of may be highlighted in the standard/basic Market Browser WARNING: This has no checks for Corporation Roles.'), ('advanced_market_browser', 'Can access the Avanced Market Browser'), ('can_add_token_character', 'Can add a Character Token with required scopes'), ('can_add_token_corporation', 'Can add a Corpration Token with required scopes'))},
        ),
    ]



# Generated by Django 4.0.8 on 2022-12-07 06:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0007_evetype_description'),
        ('marketmanager', '0041_alter_general_options_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='priceconfig',
            name='buy_order',
            field=models.BooleanField(default=False, verbose_name='Buy Order'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='eve_group',
            field=models.ManyToManyField(blank=True, to='eveuniverse.evegroup', verbose_name='EVE Inventory Group'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='eve_type',
            field=models.ManyToManyField(blank=True, to='eveuniverse.evetype', verbose_name='EVE Type'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='scalp',
            field=models.BooleanField(default=False, help_text='True for Scalp, False for Bargain', verbose_name='Scalp Finder?'),
        ),
    ]



import datetime
import random
from typing import Union

from discord import Embed
from eveuniverse.models import EveRegion, EveSolarSystem, EveStation, EveType

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F, Q, QuerySet, Sum

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from esi.models import Token

from marketmanager.app_settings import (
    MARKETMANAGER_WEBHOOK_COLOUR_INFO, MARKETMANAGER_WEBHOOK_COLOUR_WARNING,
    get_site_url,
)
from marketmanager.models import (
    Order, PriceConfig, PrivateConfig, Structure, SupplyConfig, TypeStatistics,
)

from . import __version__
from .providers import get_characters_character_id_roles_from_token

logger = get_extension_logger(__name__)


def get_corp_token(corporation_id: int, scopes: list, req_roles: list) -> Token:
    """
    Helper method to get a token for a specific character from a specific corp with specific scopes
    :param corp_id: Corp to filter on.
    :param scopes: array of ESI scope strings to search for.
    :param req_roles: roles required on the character.
    :return: :class:esi.models.Token or False
    """
    if 'esi-characters.read_corporation_roles.v1' not in scopes:
        scopes.append("esi-characters.read_corporation_roles.v1")

    char_ids = EveCharacter.objects.filter(
        corporation_id=corporation_id).values('character_id')
    tokens = Token.objects \
        .filter(character_id__in=char_ids) \
        .require_scopes(scopes)

    for token in tokens:
        roles = get_characters_character_id_roles_from_token(token)
        has_roles = False
        for role in roles.get('roles', []):
            if role in req_roles:
                has_roles = True

        if has_roles:
            return token
        else:
            pass  # TODO Maybe remove token?

    return False


def get_random_market_token() -> Union[Token, bool]:
    """Very specific edge case, we need _any_ token in order to view data on public structures.

    Args:
        scopes: array of ESI scope strings to search for.

    Returns:
        Matching token
    """
    required_scopes = ['esi-markets.structure_markets.v1']
    try:
        random_token = random.choice(
            Token.objects.all().require_scopes(
                required_scopes
            )
        )
        return random_token
    except Exception as e:
        logger.exception(e)
        return False


def is_existing_order(order, current_orders):
    try:
        existing_order = current_orders.get(order_id=order["order_id"])
        return existing_order
    except ObjectDoesNotExist:
        return False


def get_matching_privateconfig_token(structure_id: int) -> Union[Token, bool]:
    structure = Structure.objects.get(structure_id=structure_id)

    configured_tokens = PrivateConfig.objects.filter(valid_structures=structure)
    if configured_tokens.count() == 0:
        try:
            structure = Structure.objects.get(structure_id=structure_id)
            corporation = EveCorporationInfo.objects.get(
                corporation_id=structure.owner_id)
        except ObjectDoesNotExist:
            return False
        configured_tokens = PrivateConfig.objects.filter(valid_corporations=corporation)
    try:
        return random.choice(configured_tokens).token
    except IndexError:
        return False


def create_embed_base(config: Union[SupplyConfig, PriceConfig]):
    embed = Embed()
    embed.timestamp = datetime.datetime.now()
    embed.set_footer(
        icon_url="",
        text=f"AA Market Manager v{__version__}"
    )

    if config.buy_order is True:
        buy_order_string = "BUY"
    else:
        buy_order_string = "SELL"
    embed.add_field(
        name="Order",
        value=buy_order_string,
        inline=True
    )

    if config.structure.count() > 0 or config.solar_system.count() > 0 or config.region.count() > 0:
        locations_string = ""
        for structure in config.structure.all():
            locations_string = locations_string + f"{structure.name} \n"
        for solar_system in config.solar_system.all():
            locations_string = locations_string + f"{solar_system.name} \n"
        for region in config.region.all():
            locations_string = locations_string + f"{region.name} \n"
        if locations_string == "":
            locations_string = "New Eden"
        embed.add_field(
            name="Locations",
            value=locations_string,
            inline=False
        )

    if config.structure_type.count() > 0:
        structure_type_string = ""
        for type in config.structure_type.all():
            structure_type_string = structure_type_string + f"{type.name} \n"

        embed.add_field(
            name="Structure_Types",
            value=structure_type_string,
            inline=False
        )

    return embed


def create_embed_supply(
        config: SupplyConfig,
        matched_volume: int,
        description: str,
        calculated_price: int = 0,
        colour: int = MARKETMANAGER_WEBHOOK_COLOUR_INFO):

    embed = create_embed_base(config)
    embed.title = f"{config.eve_type.name}"
    embed.url = f"{get_site_url()}/marketmanager/marketbrowser?type_id={config.eve_type.id}"
    embed.set_thumbnail(url=config.eve_type.icon_url())
    embed.description = description
    embed.colour = colour
    embed.insert_field_at(
        index=2,
        name="Volume",
        value=f"{matched_volume:,} / {config.volume:,}",
        inline=True
    )
    if config.jita_compare_percent != 0:
        embed.insert_field_at(
            index=3,
            name=f"Jita {config.jita_compare_percent:,}%",
            value=f"{calculated_price:,} ISK",
            inline=True
        )
    elif config.price != 0:
        embed.insert_field_at(
            index=3,
            name="Price",
            value=f"{config.price:,} ISK",
            inline=True
        )

    if config.managed_supply_config is not None:
        if config.managed_supply_config.managed_app == 'fittings':
            managed_app = "AA-Fittings"
        else:
            managed_app = config.managed_supply_config.managed_app

        embed.add_field(
            name=managed_app,
            value=f"{config.managed_supply_config.managed_app_reason} x{config.managed_supply_config.managed_quantity}",
            inline=True
        )

    return embed


def create_embed_price(
        config: PriceConfig,
        description: str,
        calculated_price: int = 0,
        colour: int = MARKETMANAGER_WEBHOOK_COLOUR_INFO):

    embed = create_embed_base(config)
    if config.scalp is True:
        embed.title = 'Price Check: Scalping'
    else:
        embed.title = 'Price Check: Bargains'
    embed.url = f"{get_site_url()}/marketmanager/marketbrowser"
    # embed.set_thumbnail(url=config.eve_type.icon_url())
    embed.description = description
    embed.colour = MARKETMANAGER_WEBHOOK_COLOUR_WARNING

    if config.jita_compare_percent != 0:
        embed.insert_field_at(
            index=2,
            name=f"Jita {config.jita_compare_percent:,}%",
            value="\u200b",
            inline=True
        )
    elif config.price != 0:
        embed.insert_field_at(
            index=2,
            name="Price",
            value=f"{config.price:,} ISK",
            inline=True
        )

    return embed


def fifth_percentile(eve_type: EveType, eve_region: EveRegion = EveRegion(None), buy_order: bool = False):
    if eve_region.id is None:
        orders = Order.objects.filter(
            eve_type=eve_type, is_buy_order=buy_order).order_by('price')
    else:
        orders = Order.objects.filter(
            eve_type=eve_type, eve_region=eve_region, is_buy_order=buy_order).order_by('price')

    if buy_order is True:
        orders.reverse

    total_volume = orders.aggregate(volume=Sum(F('volume_remain')))['volume']
    if total_volume is None or 0:
        return 0

    cumulative_volume = 0

    for order in orders:
        cumulative_volume = cumulative_volume + order.volume_remain
        if cumulative_volume > (total_volume * 0.05):
            return order.price

    return 0


def median(eve_type: EveType, eve_region: EveRegion = EveRegion(None), buy_order: bool = False):
    # Hey median is just 50th percentile lol
    # Can we combine these?
    if eve_region.id is None:
        orders = Order.objects.filter(
            eve_type=eve_type, is_buy_order=buy_order).order_by('price')
    else:
        orders = Order.objects.filter(
            eve_type=eve_type, eve_region=eve_region, is_buy_order=buy_order).order_by('price')

    if buy_order is True:
        orders.reverse

    total_volume = orders.aggregate(volume=Sum(F('volume_remain')))['volume']
    if total_volume is None or 0:
        return 0
    cumulative_volume = 0

    for order in orders:
        cumulative_volume = cumulative_volume + order.volume_remain
        if cumulative_volume > (total_volume * 0.50):
            return order.price

    return 0


def weighted_average(eve_type: EveType, eve_region: EveRegion = EveRegion(None), buy_order: bool = False):
    if eve_region.id is None:
        orders = Order.objects.filter(eve_type=eve_type, is_buy_order=buy_order)
    else:
        orders = Order.objects.filter(
            eve_type=eve_type, eve_region=eve_region, is_buy_order=buy_order)

    total_volume = orders.aggregate(aggregate_volume=Sum(
        F('volume_remain')))['aggregate_volume']
    cumulative_price = orders.aggregate(aggregate_price=Sum(
        F('volume_remain') * F("price")))['aggregate_price']

    try:
        return cumulative_price / total_volume
    except (ZeroDivisionError, TypeError):
        return 0


def filter_orders_by_location(
        orders: QuerySet,
        config: Union[SupplyConfig, PriceConfig]
) -> QuerySet:
    if config.structure.count() > 0 or config.solar_system.count() > 0 or config.region.count() > 0:
        matching_orders_location = orders.filter(
            Q(location_id__in=config.structure.all()) | Q(eve_solar_system__in=config.solar_system.all()) | Q(eve_region__in=config.region.all()))
    else:
        matching_orders_location = orders

    if config.structure_type.count() > 0:
        valid_structures = Structure.objects.filter(
            eve_type__in=config.structure_type.all())
        matching_orders_location_type = matching_orders_location.filter(
            location_id__in=valid_structures.all())
        return matching_orders_location_type

    return matching_orders_location


def calculate_comparator_price(config: SupplyConfig):
    if config.price != 0:
        return config.price

    if config.jita_compare_percent != 0:
        try:
            type_statistics = TypeStatistics.objects.get(
                eve_type=config.eve_type,
                eve_region=EveRegion.objects.get(id=10000002))
            if config.buy_order is True:
                return type_statistics.buy_weighted_average * (config.jita_compare_percent / 100)
            if config.buy_order is False:
                return type_statistics.sell_weighted_average * (config.jita_compare_percent / 100)
        except ObjectDoesNotExist:
            return 0
    else:
        return 0


def location_resolver_to_solar_system(location_id: int) -> EveSolarSystem:
    """
    Return the Solar System for a location_id, if known

    :param location_id: _description_
    :type location_id: int
    :return: _description_
    :rtype: EveSolarSystem
    """
    if location_id >= 60000000 and location_id <= 64000000:
        try:
            station, fetched = EveStation.objects.get_or_create_esi(id=location_id)
            return station.eve_solar_system
        except Exception as e:
            logger.debug(e)
            return EveSolarSystem(None)
    else:
        try:
            return Structure.objects.get(structure_id=location_id).solar_system
        except Exception as e:
            logger.debug(e)
            return EveSolarSystem(None)



# Generated by Django 4.0.3 on 2022-03-15 13:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0012_remove_watchconfig_constellation'),
    ]

    operations = [
        migrations.AddField(
            model_name='watchconfig',
            name='managed_app',
            field=models.CharField(default='', help_text='The App managing this WatchConfig', max_length=50, verbose_name='Managed By App'),
        ),
        migrations.AddField(
            model_name='watchconfig',
            name='managed_app_identifier',
            field=models.CharField(default='', help_text='An identifier relevant to the App managing this WatchConfig', max_length=50, verbose_name='Managed App-Identifier'),
        ),
        migrations.AddField(
            model_name='watchconfig',
            name='manged_app_reason',
            field=models.CharField(default='', help_text='User Facing reason for this Managed WatchConfig', max_length=50, verbose_name='Managed App Reason'),
        ),
    ]



from eveuniverse.models import EveType

from django.test import TestCase

from allianceauth.authentication.models import EveCharacter

from marketmanager.models import Order

from .testdata.load_eveuniverse import load_eveuniverse


def create_testdata():
    load_eveuniverse()
    EveCharacter.objects.all().delete()
    EveCharacter.objects.create(
        character_id=1,
        character_name='Character1',
        corporation_id=1,
        corporation_name='test corp',
        corporation_ticker='TEST'
    )
    EveCharacter.objects.create(
        character_id=2,
        character_name='Character2',
        corporation_id=1,
        corporation_name='test corp',
        corporation_ticker='TEST'
    )
    Order.objects.create(
        # expired order
        eve_type=EveType.objects.get(id=44992)
    )


class TestMarketmanagerTasks(TestCase):
    def test_garbage_collection(self):
        # Orders
        # Generic Expiry Date Passed

        # Calculated Expiry Date Passed

        # State = Expired or Cancelled

        # Stale

        # Structures
        # Stale
        self.assertTrue(True)



import re

from django.apps import apps
from django.conf import settings


def get_site_url():  # regex sso url
    regex = r"^(.+)\/s.+"
    matches = re.finditer(regex, settings.ESI_SSO_CALLBACK_URL, re.MULTILINE)
    url = "http://"

    for m in matches:
        url = m.groups()[0]  # first match

    return url


def discord_bot_active() -> bool:
    return apps.is_installed("aadiscordbot")


def fittings_active():
    return apps.is_installed('fittings')


MARKETMANAGER_CLEANUP_DAYS_STRUCTURE = getattr(settings, "MARKETMANAGER_CLEANUP_DAYS_STRUCTURE", 7)
MARKETMANAGER_CLEANUP_DAYS_ORDER = getattr(settings, "MARKETMANAGER_CLEANUP_DAYS_ORDER", 7)
MARKETMANAGER_TASK_PRIORITY_ORDERS = getattr(settings, "MARKETMANAGER_TASK_PRIORITY_ORDERS", 5)
MARKETMANAGER_TASK_PRIORITY_STRUCTURES = getattr(settings, "MARKETMANAGER_TASK_PRIORITY_STRUCTURES", 4)
MARKETMANAGER_TASK_PRIORITY_BACKGROUND = getattr(settings, "MARKETMANAGER_TASK_PRIORITY_BACKGROUND", 7)
MARKETMANAGER_TASK_PRIORITY_SUPPLY_CONFIGS = getattr(settings, "MARKETMANAGER_TASK_PRIORITY_SUPPLY_CONFIGS", 6)
MARKETMANAGER_TASK_PRIORITY_PRICE_CONFIGS = getattr(settings, "MARKETMANAGER_TASK_PRIORITY_PRICE_CONFIGS", 6)
MARKETMANAGER_WEBHOOK_COLOUR_ERROR = getattr(settings, "MARKETMANAGER_WEBHOOK_COLOUR_ERROR", 16711710)
MARKETMANAGER_WEBHOOK_COLOUR_WARNING = getattr(settings, "MARKETMANAGER_WEBHOOK_COLOUR_WARNING", 14177041)
MARKETMANAGER_WEBHOOK_COLOUR_INFO = getattr(settings, "MARKETMANAGER_WEBHOOK_COLOUR_INFO", 42751)
MARKETMANAGER_WEBHOOK_COLOUR_SUCCESS = getattr(settings, "MARKETMANAGER_WEBHOOK_COLOUR_SUCCESS", 6684416)
MARKETMANAGER_TYPESTATISTICS_MINIMUM_ORDER_COUNT = getattr(
    settings, "MARKETMANAGER_TYPESTATISTICS_MINIMUM_ORDER_COUNT", 10)

MARKETMANAGER_COG_OPTIMIZE_FREIGHT_COST: float = getattr(settings, "MARKETMANAGER_COG_OPTIMIZE_FREIGHT_COST", 700.0)
MARKETMANAGER_COG_OPTIMIZE_REFINE_RATE: float = getattr(settings, "MARKETMANAGER_COG_OPTIMIZE_REFINE_RATE", 0.9063)



# Generated by Django 4.0.3 on 2022-03-15 13:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0011_rename_location_watchconfig_structure'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='watchconfig',
            name='constellation',
        ),
    ]



# Generated by Django 4.0.6 on 2022-10-22 06:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0039_alter_order_state'),
    ]

    operations = [
        migrations.RenameModel('ManagedWatchConfig', 'ManagedSupplyConfig'),
        migrations.RenameModel('WatchConfig', 'SupplyConfig')
    ]



from eveuniverse.models import (
    EveGroup, EveMarketGroup, EveRegion, EveSolarSystem, EveType,
)

from django.contrib import admin
from django.forms import ModelMultipleChoiceField

from allianceauth.services.hooks import get_extension_logger
from esi.models import Token

from marketmanager.app_settings import (
    MARKETMANAGER_TASK_PRIORITY_PRICE_CONFIGS,
    MARKETMANAGER_TASK_PRIORITY_STRUCTURES,
    MARKETMANAGER_TASK_PRIORITY_SUPPLY_CONFIGS, discord_bot_active,
)
from marketmanager.models import (
    Channel, ManagedSupplyConfig, Order, PriceConfig, PrivateConfig,
    PublicConfig, StatisticsConfig, Structure, SupplyConfig, Webhook,
)
from marketmanager.tasks import (
    fetch_markets_structures_structure_id, run_price_config, run_supply_config,
)

from .admin_helpers import list_2_html_w_tooltips

logger = get_extension_logger(__name__)


@admin.register(PublicConfig)
class PublicConfigAdmin(admin.ModelAdmin):
    filter_horizontal = ["fetch_regions"]

    def formfield_for_manytomany(self, db_field, request, **kwargs) -> ModelMultipleChoiceField:
        # This should filter regions to valid market regions, ie. no shattered, AT or Jove
        if db_field.name == "fetch_regions":
            kwargs["queryset"] = EveRegion.objects.filter(id__lt="11000000")
        return super().formfield_for_manytomany(db_field, request, **kwargs)


@admin.register(StatisticsConfig)
class StatisticsConfigAdmin(admin.ModelAdmin):
    filter_horizontal = ["calculate_regions"]

    def formfield_for_manytomany(self, db_field, request, **kwargs) -> ModelMultipleChoiceField:
        # This should filter regions to valid market regions, ie. no shattered, AT or Jove
        if db_field.name == "calculate_regions":
            kwargs["queryset"] = EveRegion.objects.filter(id__lt="11000000")
        return super().formfield_for_manytomany(db_field, request, **kwargs)


@admin.register(PrivateConfig)
class PrivateConfigAdmin(admin.ModelAdmin):
    list_display = ['token', 'failed', 'failure_reason']
    filter_horizontal = ["valid_corporations", "valid_structures"]

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "token":
            required_scopes = ['esi-markets.structure_markets.v1']
            kwargs["queryset"] = Token.objects.all(
            ).require_scopes(required_scopes)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)


@admin.register(SupplyConfig)
class SupplyConfigAdmin(admin.ModelAdmin):
    list_display = ['eve_type', 'buy_order', 'volume', 'price', '_locations', 'last_result_volume', 'last_run']
    filter_horizontal = ["structure", "structure_type", "webhooks", "debug_webhooks", "solar_system", "region"]
    autocomplete_fields = ['eve_type']
    list_filter = ['buy_order', 'webhooks', 'debug_webhooks', 'managed_supply_config']
    actions = ['run_selected_supply_configs']

    if discord_bot_active:
        filter_horizontal.append("channels")
        filter_horizontal.append("debug_channels")

    def formfield_for_manytomany(self, db_field, request, **kwargs) -> ModelMultipleChoiceField:
        # This should filter Citadels (1657) and NPC Stations (15) as viable Structure_Types for this selection
        # This should filter regions to valid market regions, ie. no shattered, AT or Jove
        if db_field.name == "structure_type":
            kwargs["queryset"] = EveType.objects.filter(
                eve_group__id__in=[15, 1404, 1406, 1657], published=1)
        if db_field.name == "region":
            kwargs["queryset"] = EveRegion.objects.filter(id__lt="11000000")
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    @admin.action(description="Run Selected SupplyConfigs")
    def run_selected_supply_configs(modeladmin, request, queryset):
        for config in queryset:
            run_supply_config.apply_async(
                args=[config.id], priority=MARKETMANAGER_TASK_PRIORITY_SUPPLY_CONFIGS)

    def _locations(self, obj) -> str:
        locations = [x.name for x in obj.structure.order_by(
            'name')] + [x.name for x in obj.solar_system.order_by('name')] + [x.name for x in obj.region.order_by('name')]

        return list_2_html_w_tooltips(locations, 10)


@admin.register(PriceConfig)
class PriceConfigAdmin(admin.ModelAdmin):
    list_display = ['_types', 'buy_order', 'price', '_locations', 'last_run']
    filter_horizontal = [
        'eve_type', "eve_market_group", "eve_group", "structure", "structure_type",
        "webhooks", "debug_webhooks", "solar_system", "region"]
    list_filter = ['buy_order', 'webhooks', 'debug_webhooks']
    actions = ['run_selected_price_configs']

    if discord_bot_active:
        filter_horizontal.append("channels")
        filter_horizontal.append("debug_channels")

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        # Only items that have market groups?
        if db_field.name == "eve_type":
            kwargs["queryset"] = EveType.objects.filter(
                eve_market_group__isnull=False, published=1)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_manytomany(self, db_field, request, **kwargs) -> ModelMultipleChoiceField:
        # This should filter Citadels (1657) and NPC Stations (15) as viable Structure_Types for this selection
        # This should filter regions to valid market regions, ie. no shattered, AT or Jove
        if db_field.name == "structure_type":
            kwargs["queryset"] = EveType.objects.filter(
                eve_group__id__in=[15, 1404, 1406, 1657], published=1)
        if db_field.name == "eve_region":
            kwargs["queryset"] = EveRegion.objects.filter(id__lt="11000000")
        if db_field.name == "eve_solar_system":
            kwargs["queryset"] = EveSolarSystem.objects.filter(
                id__lt="31000000")
        if db_field.name == "eve_group":
            kwargs["queryset"] = EveGroup.objects.filter(published=1)
        if db_field.name == "eve_market_group":
            kwargs["queryset"] = EveMarketGroup.objects.filter(
                parent_market_group__isnull=True)
        if db_field.name == "eve_type":
            kwargs["queryset"] = EveType.objects.filter(
                eve_market_group__isnull=False, published=1)
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    @admin.action(description="Run Selected PriceConfig")
    def run_selected_price_configs(modeladmin, request, queryset):
        for config in queryset:
            run_price_config.apply_async(
                args=[config.id], priority=MARKETMANAGER_TASK_PRIORITY_PRICE_CONFIGS)

    def _locations(self, obj):
        my_groups = [x.name for x in obj.structure.order_by('name')]

        return list_2_html_w_tooltips(my_groups, 10)

    def _types(self, obj):
        my_groups = [x.name for x in obj.eve_type.order_by('name')]

        return list_2_html_w_tooltips(my_groups, 10)


@admin.register(Order)
class OrderAdmin(admin.ModelAdmin):
    list_display = ('eve_type', 'price', 'eve_solar_system', 'is_buy_order',
                    'issued_by_character', 'issued_by_corporation', 'updated_at')


@admin.register(Structure)
class StructureAdmin(admin.ModelAdmin):
    list_display = ('name', 'solar_system', 'eve_type', 'owner_id', 'pull_market', 'updated_at')
    actions = ['pull_selected_structure_orders']
    autocomplete_fields = ['eve_type']

    @admin.action(description="Pull order from Selected Structures")
    def pull_selected_structure_orders(modeladmin, request, queryset):
        for structure in queryset:
            fetch_markets_structures_structure_id.apply_async(
                args=[structure.structure_id], priority=MARKETMANAGER_TASK_PRIORITY_STRUCTURES)


@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    list_display = ('name', )


@admin.register(Channel)
class ChannelAdmin(admin.ModelAdmin):
    list_display = ('name', )


@admin.register(ManagedSupplyConfig)
class ManagedSupplyConfigAdmin(admin.ModelAdmin):
    list_display = ['managed_app_reason', 'managed_quantity', 'managed_app_identifier']
    filter_horizontal = [
        "managed_structure", "managed_structure_type", "managed_webhooks",
        "managed_debug_webhooks", "managed_solar_system", "managed_region"]
    list_filter = ['managed_buy_order', 'managed_webhooks', 'managed_debug_webhooks']

    if discord_bot_active:
        filter_horizontal.append("managed_channels")
        filter_horizontal.append("managed_debug_channels")

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        # Only items that have market groups?
        if db_field.name == "managed_eve_type":
            kwargs["queryset"] = EveType.objects.filter(
                eve_market_group__isnull=False, published=1)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_manytomany(self, db_field, request, **kwargs) -> ModelMultipleChoiceField:
        # This should filter Citadels (1657) and NPC Stations (15) as viable Structure_Types for this selection
        # This should filter regions to valid market regions, ie. no shattered, AT or Jove
        if db_field.name == "managed_structure_type":
            kwargs["queryset"] = EveType.objects.filter(
                eve_group__id__in=[15, 1404, 1406, 1657], published=1)
        if db_field.name == "managed_region":
            kwargs["queryset"] = EveRegion.objects.filter(id__lt="11000000")
        return super().formfield_for_manytomany(db_field, request, **kwargs)



# Generated by Django 4.0.3 on 2022-03-13 14:17

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0005_type_materials_and_sections'),
        ('marketmanager', '0011_alter_privateconfig_failed_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='structure',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='structure',
            name='updated_at',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.RemoveField(
            model_name='watchconfig',
            name='type',
        ),
        migrations.AddField(
            model_name='watchconfig',
            name='type',
            field=models.ForeignKey(default='', on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype', verbose_name='EVE Type'),
            preserve_default=False,
        ),
    ]



# Generated by Django 4.0.10 on 2023-08-04 04:28

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0007_evetype_description'),
        ('marketmanager', '0046_alter_managedsupplyconfig_options_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='structure',
            name='eve_type',
            field=models.ForeignKey(limit_choices_to={'eve_group__id__in': [15, 1657], 'published': 1}, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveuniverse.evetype', verbose_name='Type'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='eve_type',
            field=models.ForeignKey(limit_choices_to={'eve_market_group__isnull': False, 'published': 1}, on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype', verbose_name='EVE Type'),
        ),
        migrations.AlterField(
            model_name='typestatistics',
            name='eve_region',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.everegion', verbose_name='Region'),
        ),
        migrations.AlterField(
            model_name='typestatistics',
            name='eve_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype', verbose_name='Type'),
        ),
    ]



from eveuniverse.tools.testdata import ModelSpec, create_testdata

from django.test import TestCase

from . import test_data_filename


class CreateEveUniverseTestData(TestCase):
    def test_create_testdata(self):
        testdata_spec = [
            ModelSpec(
                "EveType",
                ids=[44992, 22436, 621],
            ),
            ModelSpec(
                "EveSolarSystem", ids=[30000142, 30002187, 30002510],
            ),
            ModelSpec("EveRegion", ids=[10000002,10000002, 10000003], include_children=True),
        ]
        create_testdata(testdata_spec, test_data_filename())



# Generated by Django 4.0.8 on 2022-12-07 06:26

from decimal import Decimal

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0007_evetype_description'),
        ('marketmanager', '0040_rename_supplyconfigmodels'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='general',
            options={'default_permissions': (), 'managed': False, 'permissions': (('basic_market_browser', 'Can access the Standard Market Browser'), ('basic_market_watches', 'Can view all SupplyConfigs and their status'), ('order_highlight_user', "Market orders owned by the user's characters may be highlighted in the standard/basic Market Browser"), ('order_highlight_corporation', 'Market orders owned by any corporation a user is a member of may be highlighted in the standard/basic Market Browser WARNING: This has no checks for Corporation Roles.'), ('advanced_market_browser', 'Can access the Avanced Market Browser'), ('can_add_token_character', 'Can add a Character Token with required scopes'), ('can_add_token_corporation', 'Can add a Corpration Token with required scopes'))},
        ),
        migrations.RenameField(
            model_name='supplyconfig',
            old_name='managed_watch_config',
            new_name='managed_supply_config',
        ),
        migrations.RemoveField(
            model_name='supplyconfig',
            name='config_type',
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_app',
            field=models.CharField(blank=True, default='', help_text='The App managing this SupplyConfig', max_length=50, verbose_name='Managed By App'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_app_identifier',
            field=models.CharField(blank=True, default='', help_text='An identifier relevant to the App managing this SupplyConfig', max_length=50, verbose_name='Managed App-Identifier'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_app_reason',
            field=models.CharField(blank=True, default='', help_text='User Facing reason for this Managed SupplyConfig', max_length=50, verbose_name='Managed App Reason'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_debug_channels',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on SupplyConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.channel', verbose_name='Debug Channels'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_debug_webhooks',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on Supplyonfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.webhook', verbose_name='Debug Webhook'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_structure_type',
            field=models.ManyToManyField(blank=True, help_text='Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)', to='eveuniverse.evetype', verbose_name='Structure Type Filter'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='debug_channels',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on SupplyConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.channel', verbose_name='Debug Channels'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='debug_webhooks',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on SupplyConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.webhook', verbose_name='Debug Webhook'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='structure_type',
            field=models.ManyToManyField(blank=True, help_text='Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)', related_name='+', to='eveuniverse.evetype', verbose_name='Structure Type Filter'),
        ),
        migrations.CreateModel(
            name='PriceConfig',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('buy_order', models.BooleanField(verbose_name='Buy Order')),
                ('scalp', models.BooleanField(help_text='True for Scalp, False for Bargain', verbose_name='Scalp Finder?')),
                ('price', models.DecimalField(decimal_places=2, default=Decimal('0'), max_digits=15, verbose_name='Price')),
                ('jita_compare_percent', models.DecimalField(decimal_places=0, default=Decimal('0'), help_text='If set ignores Flat price value', max_digits=4, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(1000)], verbose_name='Jita Comparison %')),
                ('last_run', models.DateTimeField(blank=True, null=True, verbose_name='Last Task Runtime')),
                ('channels', models.ManyToManyField(blank=True, to='marketmanager.channel', verbose_name='Channels')),
                ('debug_channels', models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on PriceConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.channel', verbose_name='Debug Channels')),
                ('debug_webhooks', models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on PriceConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.webhook', verbose_name='Debug Webhook')),
                ('eve_group', models.ManyToManyField(to='eveuniverse.evegroup', verbose_name='EVE Inventory Group')),
                ('eve_type', models.ManyToManyField(to='eveuniverse.evetype', verbose_name='EVE Type')),
                ('region', models.ManyToManyField(blank=True, to='eveuniverse.everegion', verbose_name='Region')),
                ('solar_system', models.ManyToManyField(blank=True, to='eveuniverse.evesolarsystem', verbose_name='Solar System')),
                ('structure', models.ManyToManyField(blank=True, to='marketmanager.structure', verbose_name='Structure')),
                ('structure_type', models.ManyToManyField(blank=True, help_text='Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)', related_name='+', to='eveuniverse.evetype', verbose_name='Structure Type Filter')),
                ('webhooks', models.ManyToManyField(blank=True, to='marketmanager.webhook', verbose_name='Webhooks')),
            ],
            options={
                'default_permissions': (),
            },
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-15 14:29

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0005_type_materials_and_sections'),
        ('marketmanager', '0014_alter_watchconfig_price_typestatistics'),
    ]

    operations = [
        migrations.AddField(
            model_name='typestatistics',
            name='eve_region',
            field=models.ForeignKey(default='', on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.everegion', verbose_name=''),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='typestatistics',
            name='buy_5_percent',
            field=models.DecimalField(decimal_places=2, max_digits=5, verbose_name='Buy 5th Percentile'),
        ),
        migrations.AlterField(
            model_name='typestatistics',
            name='buy_median',
            field=models.DecimalField(decimal_places=2, max_digits=5, verbose_name='Buy Median'),
        ),
        migrations.AlterField(
            model_name='typestatistics',
            name='buy_weighted_average',
            field=models.DecimalField(decimal_places=2, max_digits=5, verbose_name='Buy Weighted Average'),
        ),
        migrations.AlterField(
            model_name='typestatistics',
            name='eve_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype', verbose_name=''),
        ),
        migrations.AlterField(
            model_name='typestatistics',
            name='sell_5_percent',
            field=models.DecimalField(decimal_places=2, max_digits=5, verbose_name='Sell 5th Percentile'),
        ),
        migrations.AlterField(
            model_name='typestatistics',
            name='sell_median',
            field=models.DecimalField(decimal_places=2, max_digits=5, verbose_name='Sell Median'),
        ),
        migrations.AlterField(
            model_name='typestatistics',
            name='sell_weighted_average',
            field=models.DecimalField(decimal_places=2, max_digits=5, verbose_name='Sell Weighted Average'),
        ),
        migrations.AddConstraint(
            model_name='typestatistics',
            constraint=models.UniqueConstraint(fields=('eve_type', 'eve_region'), name='EveTypeEveRegionTypeStatistic'),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-21 14:04

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0026_alter_publicconfig_options_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='ManagedWatchConfig',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('managed_app', models.CharField(blank=True, default='', help_text='The App managing this WatchConfig', max_length=50, verbose_name='Managed By App')),
                ('managed_app_identifier', models.CharField(blank=True, default='', help_text='An identifier relevant to the App managing this WatchConfig', max_length=50, verbose_name='Managed App-Identifier')),
                ('manged_app_reason', models.CharField(blank=True, default='', help_text='User Facing reason for this Managed WatchConfig', max_length=50, verbose_name='Managed App Reason')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.RemoveField(
            model_name='watchconfig',
            name='managed_app',
        ),
        migrations.RemoveField(
            model_name='watchconfig',
            name='managed_app_identifier',
        ),
        migrations.RemoveField(
            model_name='watchconfig',
            name='manged_app_reason',
        ),
        migrations.AddField(
            model_name='watchconfig',
            name='managed_watch_config',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='marketmanager.managedwatchconfig', verbose_name=''),
        ),
    ]



# Generated by Django 4.2.6 on 2023-11-03 13:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0050_alter_managedsupplyconfig_managed_channels_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='privateconfig',
            name='failure_reason',
            field=models.CharField(blank=True, default='Failure Reason', max_length=100, verbose_name='Failure Reason'),
        ),
    ]



# import json
# from unittest import mock

# from django.test import TestCase

# from marketmanager.models import Order, Structure
# from marketmanager.providers import (
#     get_characters_character_id_orders, get_corporations_corporation_id_orders,
#     get_corporations_corporation_id_structures,
#     get_markets_region_id_orders_paged, get_markets_structures_structure_id,
#     get_universe_structures, get_universe_structures_structure_id,
# )



# Generated by Django 3.2.8 on 2021-11-30 12:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0003_alter_general_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='order',
            name='wallet_division',
            field=models.PositiveSmallIntegerField(blank=True, help_text='The corporation wallet division used for this order.', null=True, verbose_name='Wallet Division'),
        ),
    ]



# Generated by Django 4.0.8 on 2022-12-07 12:42

from decimal import Decimal

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0007_evetype_description'),
        ('marketmanager', '0042_alter_priceconfig_buy_order_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='priceconfig',
            name='eve_market_group',
            field=models.ManyToManyField(blank=True, to='eveuniverse.evemarketgroup', verbose_name='EVE Market Group'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='price',
            field=models.DecimalField(decimal_places=2, default=Decimal('0'), help_text='WARNING: Use Jita% when possible, this sets a flat price for ALL items included in this Config. Possibly useful for Caps/Soups', max_digits=15, verbose_name='Price'),
        ),
    ]



# https://developers.google.com/optimization/lp/stigler_diet
# https://github.com/cdanis/kantorotanium/blob/master/kantorotanium/__main__.py
import logging
from math import ceil

from discord.commands import SlashCommandGroup
from discord.ext import commands
from eveuniverse.models.sde import EveTypeMaterial
from eveuniverse.models.universe_1 import EveType
from ortools.linear_solver import pywraplp

from marketmanager.app_settings import (
    MARKETMANAGER_COG_OPTIMIZE_FREIGHT_COST,
    MARKETMANAGER_COG_OPTIMIZE_REFINE_RATE, MARKETMANAGER_TASK_PRIORITY_ORDERS,
)
from marketmanager.management.commands.marketmanager_preload_common_eve_types import (
    ASTEROID,
)
from marketmanager.models import Order
from marketmanager.tasks import fetch_markets_region_id_orders

logger = logging.getLogger(__name__)


class MarketManager(commands.Cog):
    """
    Market Management and Tools
    From AA-MarketManager
    """

    def __init__(self, bot):
        self.bot = bot

    marketmanager_commands = SlashCommandGroup("marketmanager", "Market Manager",)

    @marketmanager_commands.command(name="update_orders", description="Return an optimized set of orders for given minerals")
    async def update_publicmarketdata_ores(
            self, ctx, region_id: int = 10000002):
        for eve_type in EveType.objects.filter(eve_type__eve_group__eve_category_id=ASTEROID, eve_type__eve_market_group__isnull=False, eve_type__published=1):
            fetch_markets_region_id_orders.apply_async(
                args=[region_id, "sell", eve_type.id], priority=MARKETMANAGER_TASK_PRIORITY_ORDERS)
        return await ctx.respond("Tasks Queued, Please wait atleast 5 Minutes")

    @marketmanager_commands.command(name="optimize", description="Return an optimized set of orders for given minerals")
    async def optimize(
            self, ctx,
            tritanium: int, pyerite: int, mexallon: int, isogen: int, nocxium: int, zydrine: int, megacyte: int, morphite: int = 0,
            freight_cost: float = MARKETMANAGER_COG_OPTIMIZE_FREIGHT_COST,
            location_id: int = 60003760,
            raw_data: bool = False):

        output_string = []
        refine_cache = {}
        ore_minerals = []

        # Initiate the array, this is used a lot later to iterate over each mineral
        required_minerals = [
            ["Tritanium", tritanium],
            ["Pyerite", pyerite],
            ["Mexallon", mexallon],
            ["Isogen", isogen],
            ["Nocxium", nocxium],
            ["Zydrine", zydrine],
            ["Megacyte", megacyte],
            ["Morphite", morphite]]

        # Build an array of refining data to avoid ever querying it again.
        # I couldn't find a nice way to do these reverse FK's etc in the order query
        # {'Tritanium': {{type_id: quantity}, {}}, 'Mineral': ... }
        # Side note why the fuck did i do it from material upwards, i could have done Type down????
        for mineral in required_minerals:
            mineral_refine_cache = {}
            for _mineral_refine_cache in EveTypeMaterial.objects.filter(
                    eve_type__eve_group__eve_category_id=ASTEROID,
                    eve_type__eve_market_group__isnull=False, eve_type__published=1,
                    material_eve_type__name__icontains=mineral[0]).values("eve_type", "quantity", "eve_type__portion_size"):
                mineral_refine_cache.update(
                    {_mineral_refine_cache["eve_type"]: _mineral_refine_cache["quantity"] / _mineral_refine_cache["eve_type__portion_size"] * MARKETMANAGER_COG_OPTIMIZE_REFINE_RATE})
            refine_cache.update({mineral[0]: dict(mineral_refine_cache)})

        # Here we build out an array of every ore on the market, its price and what it refines into
        # Uses the above cache heavily, the market orders are quite simple
        for order in Order.objects.filter(
                eve_type__eve_group__eve_category_id=ASTEROID,
                eve_type__eve_market_group__isnull=False, eve_type__published=1, is_buy_order=0,
                location_id=location_id).select_related("eve_type"):
            # id like to replace this with a single iterator
            try:
                _tritanium = refine_cache["Tritanium"][order.eve_type.id]
            except KeyError:
                _tritanium = 0
            try:
                _pyerite = refine_cache["Pyerite"][order.eve_type.id]
            except KeyError:
                _pyerite = 0
            try:
                _mexallon = refine_cache["Mexallon"][order.eve_type.id]
            except KeyError:
                _mexallon = 0
            try:
                _isogen = refine_cache["Isogen"][order.eve_type.id]
            except KeyError:
                _isogen = 0
            try:
                _nocxium = refine_cache["Nocxium"][order.eve_type.id]
            except KeyError:
                _nocxium = 0
            try:
                _zydrine = refine_cache["Zydrine"][order.eve_type.id]
            except KeyError:
                _zydrine = 0
            try:
                _megacyte = refine_cache["Megacyte"][order.eve_type.id]
            except KeyError:
                _megacyte = 0
            try:
                _morphite = refine_cache["Morphite"][order.eve_type.id]
            except KeyError:
                _morphite = 0
            _ore_minerals = [
                f"{order.eve_type.name} @ {order.price}",
                order.eve_type.volume,
                float(order.price) + (freight_cost * order.eve_type.volume),
                order.volume_remain,
                _tritanium,
                _pyerite,
                _mexallon,
                _isogen,
                _nocxium,
                _zydrine,
                _megacyte,
                _morphite,
            ]
            ore_minerals.append(_ore_minerals)
        solver = pywraplp.Solver.CreateSolver("GLOP_LINEAR_PROGRAMMING")

        # Define all our variables in an array
        ores = [solver.NumVar(0, item[3], item[0]) for item in ore_minerals]

        # Create the constraints (Requirement), one per mineral.
        # Add coefficients/variable (Refine value), each ore ^ once per mineral
        constraints = []
        for i, mineral in enumerate(required_minerals):
            constraints.append(solver.Constraint(mineral[1], solver.infinity()))
            for j, item in enumerate(ore_minerals):
                constraints[i].SetCoefficient(ores[j], item[i + 4])

        # Objective function: Minimize the sum of price of ores.
        # This should also set the minimum viable price to one item
        # not sure if/how this affects steps
        objective = solver.Objective()
        for k, ore in enumerate(ores):
            objective.SetCoefficient(ore, ore_minerals[k][2])
        objective.SetMinimization()

        output_string.append("")
        output_string.append(
            f"Solver: {solver.SolverVersion()} Variables: {solver.NumVariables()} Constraints: {solver.NumConstraints()} ")

        status = solver.Solve()
        if status != solver.OPTIMAL:
            output_string.append("The problem does not have an optimal solution!")
            if status == solver.FEASIBLE:
                output_string.append("A potentially suboptimal solution was found.")
            else:
                output_string.append("The solver could not solve the problem.")
                return await ctx.respond("\n".join(output_string))
        required_minerals_result = [0] * len(required_minerals)

        if raw_data is False:
            output_string.append("### Required Ores:")
            output_string.append("```")
            for i, ore in enumerate(ores):
                if ore.solution_value() > 0.0:
                    output_string.append(f"{ore_minerals[i][0]} * {ceil(ore.solution_value())}")
                    for j, _ in enumerate(required_minerals):
                        required_minerals_result[j] += ore_minerals[i][j + 4] * \
                            ore.solution_value()  # i think this sums up minerals to output later
        else:
            output_string.append("### Buy All:")
            output_string.append("```")
            for i, ore in enumerate(ores):
                if ore.solution_value() > 0.0:
                    output_string.append(f"{ore_minerals[i][0].split('@')[0]} {ceil(ore.solution_value())}")
                    for j, _ in enumerate(required_minerals):
                        required_minerals_result[j] += ore_minerals[i][j + 4] * \
                            ore.solution_value()  # i think this sums up minerals to output later
        output_string.append("``` ")

        await ctx.respond("\n".join(output_string))
        output_string = []

        if raw_data is False:
            output_string.append("### Mineral Status:")
            output_string.append("```")
            for i, mineral in enumerate(required_minerals):
                output_string.append(
                    f"{mineral[0]} {required_minerals_result[i]:,.2f} / {mineral[1]} ({required_minerals_result[i] / mineral[1] * 100:.1f}%) ")
        else:
            output_string.append("### Output Minerals:")
            output_string.append("```")
            for i, mineral in enumerate(required_minerals):
                output_string.append(f"{mineral[0]} {required_minerals_result[i]:,.2f}")
        output_string.append("```")

        output_string.append(f"Cost for Requested Minerals: ISK `{objective.Value():,.2f}`")
        output_string.append("Calculation Stats:")
        output_string.append(f"Problem solved in {solver.wall_time():d} milliseconds")
        output_string.append(f"Problem solved in {solver.iterations():d} iterations")

        return await ctx.respond("\n".join(output_string))


def setup(bot):
    bot.add_cog(MarketManager(bot))



# Generated by Django 4.0.3 on 2022-03-15 04:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0008_alter_general_options'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='watchconfig',
            name='fail_webhooks',
        ),
        migrations.RemoveField(
            model_name='watchconfig',
            name='success_webhooks',
        ),
        migrations.AddField(
            model_name='watchconfig',
            name='debug_webhooks',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on WatchConfigs that dont have any issues', related_name='debug_webhooks', to='marketmanager.webhook', verbose_name='Debug Webhook'),
        ),
        migrations.AddField(
            model_name='watchconfig',
            name='webhooks',
            field=models.ManyToManyField(blank=True, to='marketmanager.webhook', verbose_name='Webhooks'),
        ),
    ]



from datetime import datetime

from eveuniverse.models import EveRegion, EveSolarSystem, EveType

from django.test import TestCase

from allianceauth.tests.auth_utils import AuthUtils
from esi.models import Token

from marketmanager.models import Order
from marketmanager.task_helpers import (
    get_corp_token, get_random_market_token, is_existing_order,
)
from marketmanager.tests.esi.utils import _generate_token, _store_as_Token

from .testdata.load_eveuniverse import load_eveuniverse


class TestTaskHelpers(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('User1')
        AuthUtils.add_main_character(
            cls.user, 'Character1', '1', corp_id='2',
            corp_name='test_corp', corp_ticker='TEST', alliance_id='3', alliance_name='TEST')
        cls.token_1 = _store_as_Token(
            _generate_token(
                character_id=1,
                character_name="Character1",
                scopes=['esi-markets.structure_markets.v1']
            ),
            cls.user
        )

    def test_get_corp_token(self):
        scopes = ["esi-markets.read_corporation_orders.v1"]
        req_roles = ["Accountant", "Trader"]
        corporation_id = 1
        token = get_corp_token(corporation_id, scopes, req_roles)

        self.assertFalse(token)

    def test_get_random_market_token(self):
        token = get_random_market_token()
        self.assertIsInstance(token, Token)

    def test_get_random_market_token_no_tokens(self):
        Token.objects.all().delete()
        token = get_random_market_token()
        self.assertIsInstance(token, bool)
        self.assertFalse(token)


class TestTaskHelpersOrders(TestCase):
    @classmethod
    def setUpTestData(cls):
        load_eveuniverse()
        Order.objects.create(
            order_id="1",
            eve_type=EveType.objects.get(id=44992),
            volume_remain=100,
            price=3000000,
            duration=90,
            location_id=60003760,
            eve_region=EveRegion.objects.get(id=10000002),
            eve_solar_system=EveSolarSystem.objects.get(id=30000142),
            issued=datetime.now(),
        )
        Order.objects.create(
            order_id="2",
            eve_type=EveType.objects.get(id=44992),
            volume_remain=100,
            price=3000000,
            duration=90,
            location_id=60003760,
            eve_region=EveRegion.objects.get(id=10000002),
            eve_solar_system=EveSolarSystem.objects.get(id=30000142),
            issued=datetime.now(),
        )

    def test_is_existing_order_existing(self):
        current_orders = Order.objects.all()
        order = {}
        order["order_id"] = 1
        self.assertTrue(is_existing_order(order, current_orders))

    def test_is_existing_order_not(self):
        current_orders = Order.objects.all()
        order = {}
        order["order_id"] = 3
        self.assertFalse(is_existing_order(order, current_orders))



# Generated by Django 4.0.3 on 2022-03-13 15:54

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    replaces = [('marketmanager', '0001_initial'), ('marketmanager', '0002_auto_20211129_1339'), ('marketmanager', '0003_alter_general_options'), ('marketmanager', '0004_alter_order_wallet_division'), ('marketmanager', '0005_alter_general_options'), ('marketmanager', '0006_auto_20211201_1343'), ('marketmanager', '0007_alter_order_is_buy_order'), ('marketmanager', '0008_alter_order_state'), ('marketmanager', '0009_rename_config_publicconfig_order_source_character_and_more'), ('marketmanager', '0010_order_source_private'), ('marketmanager', '0011_alter_privateconfig_failed_and_more'), ('marketmanager', '0012_structure_created_at_structure_updated_at_and_more'), ('marketmanager', '0013_alter_privateconfig_failure_reason_and_more'), ('marketmanager', '0014_alter_privateconfig_valid_corporations_and_more')]

    initial = True

    dependencies = [
        ('eveonline', '0015_factions'),
        ('eveuniverse', '0005_type_materials_and_sections'),
        ('esi', '0010_set_new_tokens_to_sso_v2'),
    ]

    operations = [
        migrations.CreateModel(
            name='Structure',
            fields=[
                ('structure_id', models.PositiveBigIntegerField(primary_key=True, serialize=False, verbose_name='Structure ID')),
                ('name', models.CharField(max_length=100, verbose_name='Name')),
                ('owner_id', models.IntegerField(verbose_name='Owner Corporation ID')),
                ('pull_market', models.BooleanField(default=True, help_text='Useful to ignore specific structures for _reasons_', verbose_name='Pull Market Orders')),
                ('eve_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype', verbose_name='Type')),
                ('solar_system', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evesolarsystem', verbose_name='Solar System')),
                ('created_at', models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Webhook',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, verbose_name='Name')),
                ('webhook', models.URLField(verbose_name='URL')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='General',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'permissions': (('basic_market_browser', 'Can access the Market Browser with standard features'), ('order_highlight_user', "Market orders owned by the user's characters may be highlighted in the standard/basic Market Browser"), ('order_highlight_corporation', 'Market orders owned by any corporation a user is a member of may be highlighted in the standard/basic Market Browser WARNING: This has no checks for Corporation Roles.'), ('advanced_market_browser', 'Can access the Market Browser with more detail such as the owners of orders')),
                'managed': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='PublicConfig',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('fetch_regions', models.ManyToManyField(to='eveuniverse.everegion', verbose_name='Fetch Regions')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Order',
            fields=[
                ('order_id', models.PositiveBigIntegerField(help_text='Unique order ID', primary_key=True, serialize=False, verbose_name='Order ID')),
                ('duration', models.PositiveIntegerField(help_text='Number of days the order was valid for (starting from the issued date). An order expires at time issued + duration', verbose_name='Duration')),
                ('is_buy_order', models.BooleanField(db_index=True, default=False, help_text='True if the order is a bid (buy) order', verbose_name='Buy Order')),
                ('issued', models.DateTimeField(help_text='Date and time when this order was issued', verbose_name='Issued')),
                ('location_id', models.PositiveBigIntegerField(help_text='ID of the location where order was placed', verbose_name='Location ID')),
                ('min_volume', models.PositiveIntegerField(blank=True, help_text='For buy orders, the minimum quantity that will be accepted in a matching sell order', null=True, verbose_name='Minimum Volume')),
                ('price', models.DecimalField(decimal_places=2, help_text='Cost per unit for this order', max_digits=20, verbose_name='Price')),
                ('escrow', models.DecimalField(blank=True, decimal_places=2, help_text='For buy orders, the amount of ISK in escrow', max_digits=20, null=True, verbose_name='Escrow')),
                ('range', models.CharField(choices=[('1', '1'), ('10', '10'), ('2', '2'), ('20', '20'), ('3', '3'), ('30', '30'), ('4', '4'), ('40', '40'), ('5', '5'), ('region', 'Region'), ('solarsystem', 'Solar System'), ('station', 'Station')], help_text='Valid order range, numbers are ranges in jumps', max_length=20, verbose_name='Order Range')),
                ('volume_remain', models.PositiveIntegerField(help_text='Quantity of items still required or offered', verbose_name='Volume Remaining')),
                ('volume_total', models.PositiveIntegerField(help_text='Quantity of items required or offered at time order was placed', verbose_name='Volume Total')),
                ('is_corporation', models.BooleanField(default=False, help_text='Signifies whether the buy/sell order was placed on behalf of a corporation.', verbose_name='Is Corporation')),
                ('wallet_division', models.PositiveSmallIntegerField(blank=True, help_text='The corporation wallet division used for this order.', null=True, verbose_name='Wallet Division')),
                ('state', models.CharField(choices=[('', ''), ('cancelled', 'Cancelled'), ('expired ', 'Expired')], help_text='Current order state, Only valid for Authenticated order History. Will not update from Public Market Data.', max_length=20, verbose_name='Order State')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('eve_region', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.everegion', verbose_name='Region')),
                ('eve_solar_system', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evesolarsystem', verbose_name='System')),
                ('eve_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype', verbose_name='Type')),
                ('issued_by_character', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='eveonline.evecharacter', verbose_name='Character')),
                ('issued_by_corporation', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='eveonline.evecorporationinfo', verbose_name='Corporation')),
                ('source_character', models.BooleanField(default=False, verbose_name='Sourced From Character API Market Data')),
                ('source_corporation', models.BooleanField(default=False, verbose_name='Sourced From Corporation API Market Data')),
                ('source_public', models.BooleanField(default=False, verbose_name='Sourced From Public Market Data')),
                ('source_private', models.BooleanField(default=False, verbose_name='Sourced From Public Market Data')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='WatchConfig',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('buy_order', models.BooleanField(verbose_name='Buy Order')),
                ('volume', models.IntegerField(help_text='Set to Zero to check ANY/EVERY order against Price', verbose_name='Volume')),
                ('price', models.IntegerField(help_text='Set to Zero to check ANY/EVERY order against Volume', verbose_name='Price')),
                ('jita_compare_percent', models.IntegerField(help_text='If set ignores Flat price value', verbose_name='Jita Comparison %')),
                ('constellation', models.ManyToManyField(to='eveuniverse.eveconstellation', verbose_name='Constellation')),
                ('location', models.ManyToManyField(to='marketmanager.structure', verbose_name='Structure')),
                ('region', models.ManyToManyField(to='eveuniverse.everegion', verbose_name='Region')),
                ('solar_system', models.ManyToManyField(to='eveuniverse.evesolarsystem', verbose_name='Solar System')),
                ('structure_type', models.ManyToManyField(help_text='Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)', related_name='structure_type', to='eveuniverse.evetype', verbose_name='Structure Type Filter')),
                ('webhooks', models.ManyToManyField(to='marketmanager.webhook', verbose_name='Webhooks')),
                ('type', models.ForeignKey(default='', on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype', verbose_name='EVE Type')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='PrivateConfig',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('failed', models.BooleanField(verbose_name='Disabled due to Failure, Check reason, adjust config and re-enable')),
                ('failure_reason', models.CharField(blank=True, default='', max_length=100, verbose_name='Failure Reason')),
                ('token', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='esi.token', verbose_name='ESI Token')),
                ('valid_corporations', models.ManyToManyField(blank=True, to='eveonline.evecorporationinfo', verbose_name='Valid Corporation Markets for this Token')),
                ('valid_structures', models.ManyToManyField(blank=True, to='marketmanager.structure', verbose_name='Valid Structure Markets for this Token')),
            ],
            options={
                'default_permissions': (),
            },
        ),
    ]



# Generated by Django 4.0.8 on 2023-01-30 12:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0045_alter_priceconfig_minimum'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='managedsupplyconfig',
            options={'default_permissions': (), 'verbose_name': 'Managed Supply Config'},
        ),
        migrations.AlterModelOptions(
            name='privateconfig',
            options={'default_permissions': (), 'verbose_name': 'Private Market Configuration'},
        ),
    ]



# Generated by Django 4.0.6 on 2022-08-02 11:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0030_channel_managedwatchconfig_managed_channels_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='channel',
            name='name',
            field=models.CharField(max_length=100, verbose_name='Name'),
        ),
    ]



from eveuniverse.models import EveType

from django.db.models import Sum

from allianceauth.services.hooks import get_extension_logger

from marketmanager.app_settings import fittings_active
from marketmanager.models import ManagedSupplyConfig, SupplyConfig

logger = get_extension_logger(__name__)

if fittings_active():
    from fittings.models import Fitting, FittingItem

    def update_managed_supply_config_fittings_fit(managed_supply_config: ManagedSupplyConfig):
        try:
            existing_configs = SupplyConfig.objects.filter(
                managed_supply_config=managed_supply_config)
            existing_configs.delete()
        except Exception as e:
            logger.exception(e)

        try:
            fit = Fitting.objects.get(id=managed_supply_config.managed_app_identifier.split(".")[-1])
        except Exception as e:
            logger.exception(e)

        # Ship Hull
        eve_type, eve_type_fetched = EveType.objects.get_or_create(id=fit.ship_type_type_id)
        watch_config = SupplyConfig(
            buy_order=False,
            eve_type=eve_type,
            volume=managed_supply_config.managed_quantity,
            jita_compare_percent=managed_supply_config.managed_jita_compare_percent,
            managed_supply_config=managed_supply_config
        )
        watch_config.save()
        watch_config.structure.set(managed_supply_config.managed_structure.all())
        watch_config.solar_system.set(managed_supply_config.managed_solar_system.all())
        watch_config.region.set(managed_supply_config.managed_region.all())
        watch_config.structure_type.set(managed_supply_config.managed_structure_type.all())
        watch_config.webhooks.set(managed_supply_config.managed_webhooks.all())
        watch_config.channels.set(managed_supply_config.managed_channels.all())
        watch_config.debug_webhooks.set(managed_supply_config.managed_debug_webhooks.all())
        watch_config.debug_channels.set(managed_supply_config.managed_debug_channels.all())
        watch_config.save

        # Ship Fittings
        for fit_item in FittingItem.objects.filter(fit=fit).values('type_id').annotate(quantity=Sum('quantity')):
            eve_type, eve_type_fetched = EveType.objects.get_or_create(id=fit_item['type_id'])
            watch_config = SupplyConfig(
                buy_order=False,
                eve_type=eve_type,
                volume=fit_item['quantity'] * managed_supply_config.managed_quantity,
                jita_compare_percent=managed_supply_config.managed_jita_compare_percent,
                managed_supply_config=managed_supply_config
            )
            watch_config.save()
            watch_config.structure.set(managed_supply_config.managed_structure.all())
            watch_config.solar_system.set(managed_supply_config.managed_solar_system.all())
            watch_config.region.set(managed_supply_config.managed_region.all())
            watch_config.structure_type.set(managed_supply_config.managed_structure_type.all())
            watch_config.webhooks.set(managed_supply_config.managed_webhooks.all())
            watch_config.channels.set(managed_supply_config.managed_channels.all())
            watch_config.debug_webhooks.set(managed_supply_config.managed_debug_webhooks.all())
            watch_config.debug_channels.set(managed_supply_config.managed_debug_channels.all())
            watch_config.save

    def update_managed_supply_config_fittings_doctrine():
        # lets allow doctrine selections too
        # each update_managed_supply_config_fittings_fit
        return True



# Generated by Django 4.0 on 2023-10-31 14:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0010_alter_eveindustryactivityduration_eve_type_and_more'),
        ('eveonline', '0017_alliance_and_corp_names_are_not_unique'),
        ('marketmanager', '0049_alter_general_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_channels',
            field=models.ManyToManyField(blank=True, to='marketmanager.Channel', verbose_name='Channels'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_debug_channels',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on SupplyConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.Channel', verbose_name='Debug Channels'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_debug_webhooks',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on Supplyonfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.Webhook', verbose_name='Debug Webhook'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_region',
            field=models.ManyToManyField(blank=True, to='eveuniverse.EveRegion', verbose_name='Region'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_solar_system',
            field=models.ManyToManyField(blank=True, to='eveuniverse.EveSolarSystem', verbose_name='Solar System'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_structure',
            field=models.ManyToManyField(blank=True, to='marketmanager.Structure', verbose_name='Structure'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_structure_type',
            field=models.ManyToManyField(blank=True, help_text='Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)', to='eveuniverse.EveType', verbose_name='Structure Type Filter'),
        ),
        migrations.AlterField(
            model_name='managedsupplyconfig',
            name='managed_webhooks',
            field=models.ManyToManyField(blank=True, to='marketmanager.Webhook', verbose_name='Webhooks'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='channels',
            field=models.ManyToManyField(blank=True, to='marketmanager.Channel', verbose_name='Channels'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='debug_channels',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on PriceConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.Channel', verbose_name='Debug Channels'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='debug_webhooks',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on PriceConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.Webhook', verbose_name='Debug Webhook'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='eve_group',
            field=models.ManyToManyField(blank=True, to='eveuniverse.EveGroup', verbose_name='EVE Inventory Group'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='eve_market_group',
            field=models.ManyToManyField(blank=True, to='eveuniverse.EveMarketGroup', verbose_name='EVE Market Group'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='eve_type',
            field=models.ManyToManyField(blank=True, to='eveuniverse.EveType', verbose_name='EVE Type'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='region',
            field=models.ManyToManyField(blank=True, to='eveuniverse.EveRegion', verbose_name='Region'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='solar_system',
            field=models.ManyToManyField(blank=True, to='eveuniverse.EveSolarSystem', verbose_name='Solar System'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='structure',
            field=models.ManyToManyField(blank=True, to='marketmanager.Structure', verbose_name='Structure'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='structure_type',
            field=models.ManyToManyField(blank=True, help_text='Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)', related_name='+', to='eveuniverse.EveType', verbose_name='Structure Type Filter'),
        ),
        migrations.AlterField(
            model_name='priceconfig',
            name='webhooks',
            field=models.ManyToManyField(blank=True, to='marketmanager.Webhook', verbose_name='Webhooks'),
        ),
        migrations.AlterField(
            model_name='privateconfig',
            name='valid_corporations',
            field=models.ManyToManyField(blank=True, to='eveonline.EveCorporationInfo', verbose_name='Valid Corporation Markets for this Token'),
        ),
        migrations.AlterField(
            model_name='privateconfig',
            name='valid_structures',
            field=models.ManyToManyField(blank=True, to='marketmanager.Structure', verbose_name='Valid Structure Markets for this Token'),
        ),
        migrations.AlterField(
            model_name='publicconfig',
            name='fetch_regions',
            field=models.ManyToManyField(to='eveuniverse.EveRegion', verbose_name='Fetch Regions'),
        ),
        migrations.AlterField(
            model_name='statisticsconfig',
            name='calculate_regions',
            field=models.ManyToManyField(to='eveuniverse.EveRegion', verbose_name='TypeStatistics Calculation Regions'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='channels',
            field=models.ManyToManyField(blank=True, to='marketmanager.Channel', verbose_name='Channels'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='debug_channels',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on SupplyConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.Channel', verbose_name='Debug Channels'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='debug_webhooks',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on SupplyConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='+', to='marketmanager.Webhook', verbose_name='Debug Webhook'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='region',
            field=models.ManyToManyField(blank=True, to='eveuniverse.EveRegion', verbose_name='Region'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='solar_system',
            field=models.ManyToManyField(blank=True, to='eveuniverse.EveSolarSystem', verbose_name='Solar System'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='structure',
            field=models.ManyToManyField(blank=True, to='marketmanager.Structure', verbose_name='Structure'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='structure_type',
            field=models.ManyToManyField(blank=True, help_text='Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)', related_name='+', to='eveuniverse.EveType', verbose_name='Structure Type Filter'),
        ),
        migrations.AlterField(
            model_name='supplyconfig',
            name='webhooks',
            field=models.ManyToManyField(blank=True, to='marketmanager.Webhook', verbose_name='Webhooks'),
        ),
    ]



# Generated by Django 4.0.6 on 2022-08-02 11:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0029_managedwatchconfig_managed_buy_order_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='Channel',
            fields=[
                ('name', models.CharField(max_length=100)),
                ('id', models.BigIntegerField(primary_key=True, serialize=False)),
            ],
            options={
                'verbose_name': 'Channel',
                'verbose_name_plural': 'Channels',
                'default_permissions': (),
            },
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_channels',
            field=models.ManyToManyField(blank=True, to='marketmanager.channel', verbose_name='Channels'),
        ),
        migrations.AddField(
            model_name='managedwatchconfig',
            name='managed_debug_channels',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on WatchConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='managed_debug_channels', to='marketmanager.channel', verbose_name='Debug Channels'),
        ),
        migrations.AddField(
            model_name='watchconfig',
            name='channels',
            field=models.ManyToManyField(blank=True, to='marketmanager.channel', verbose_name='Channels'),
        ),
        migrations.AddField(
            model_name='watchconfig',
            name='debug_channels',
            field=models.ManyToManyField(blank=True, help_text='Primarily for Testing/Debugging Purposes. This webhook will receive updates on WatchConfigs that _dont_ notify. Because their Configs didnt Meet/Breach.', related_name='debug_channels', to='marketmanager.channel', verbose_name='Debug Channels'),
        ),
    ]



# Generated by Django 4.0.6 on 2022-09-16 14:21

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0007_evetype_description'),
        ('marketmanager', '0037_alter_general_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='structure',
            name='eve_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveuniverse.evetype', verbose_name='Type'),
        ),
        migrations.AlterField(
            model_name='structure',
            name='solar_system',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveuniverse.evesolarsystem', verbose_name='Solar System'),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-19 06:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0021_typestatistics_new_eden'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='typestatistics',
            name='EveTypeEveRegionTypeStatistic',
        ),
    ]



# Generated by Django 4.0.10 on 2023-10-19 06:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0047_alter_structure_eve_type_alter_supplyconfig_eve_type_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='channel',
            options={'verbose_name': 'Channel', 'verbose_name_plural': 'Channels'},
        ),
        migrations.AlterModelOptions(
            name='managedsupplyconfig',
            options={'verbose_name': 'Managed Supply Config'},
        ),
        migrations.AlterModelOptions(
            name='priceconfig',
            options={},
        ),
        migrations.AlterModelOptions(
            name='privateconfig',
            options={'verbose_name': 'Private Market Configuration'},
        ),
        migrations.AlterModelOptions(
            name='publicconfig',
            options={'verbose_name': 'Public Market Configuration'},
        ),
        migrations.AlterModelOptions(
            name='statisticsconfig',
            options={'verbose_name': 'TypeStatistics Calculation Configuration'},
        ),
        migrations.AlterModelOptions(
            name='structure',
            options={},
        ),
        migrations.AlterModelOptions(
            name='supplyconfig',
            options={},
        ),
        migrations.AlterModelOptions(
            name='webhook',
            options={},
        ),
    ]



from django.urls import path

from . import views

app_name = "marketmanager"

urlpatterns = [
    path('', views.marketbrowser, name="index"),
    path('marketbrowser', views.marketbrowser, name="marketbrowser"),
    path('marketmanager', views.marketbrowser, name="marketmanager"),
    path('marketwatches', views.marketwatches, name="marketwatches"),
    path(
        'marketbrowser/ajax/search',
        views.marketbrowser_autocomplete,
        name="marketbrowser_autocomplete"
    ),
    path(
        'marketbrowser/ajax/buy_orders',
        views.marketbrowser_buy_orders,
        name="marketbrowser_buy_orders"
    ),
    path(
        'marketbrowser/ajax/sell_orders',
        views.marketbrowser_sell_orders,
        name="marketbrowser_sell_orders"
    ),
    path('char/add', views.add_char, name="add_char"),
    path('corp/add', views.add_corp, name="add_corp")
]



import datetime
from decimal import Decimal

from discord import SyncWebhook
from eveuniverse.models import (
    EveGroup, EveMarketGroup, EveRegion, EveSolarSystem, EveType,
)
from solo.models import SingletonModel

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from esi.models import Token


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_market_browser", "Can access the Standard Market Browser"),
            ("basic_market_watches", "Can view all SupplyConfigs and their status"),
            ("order_highlight_user", "Market orders owned by the user's characters may be highlighted in the standard/basic Market Browser"),
            ("order_highlight_corporation", "Market orders owned by any corporation a user is a member of may be highlighted in the standard/basic Market Browser WARNING: This has no checks for Corporation Roles."),
            ("advanced_market_browser", "Can access the Advanced Market Browser"),
            ("can_add_token_character", "Can add a Character Token with required scopes"),
            ("can_add_token_corporation", "Can add a Corpration Token with required scopes"),
        )


class Order(models.Model):
    """An EVE Market Order"""
    order_id = models.PositiveBigIntegerField(
        _("Order ID"),
        help_text=_("Unique order ID"),
        primary_key=True)
    eve_type = models.ForeignKey(
        EveType,
        verbose_name=_("Type"),
        on_delete=models.CASCADE)
    duration = models.PositiveSmallIntegerField(
        _("Duration"),
        help_text=_("Number of days the order was valid for (starting from the issued date). An order expires at time issued + duration"))
    is_buy_order = models.BooleanField(
        _("Buy Order"),
        default=False,
        help_text=_("True if the order is a bid (buy) order"),
        db_index=True)
    issued = models.DateTimeField(
        _("Issued"),
        help_text=_("Date and time when this order was issued"),
        auto_now=False,
        auto_now_add=False)
    location_id = models.PositiveBigIntegerField(
        _("Location ID"),
        help_text=_("ID of the location where order was placed"))
    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        verbose_name=_("System"),
        on_delete=models.CASCADE,
        null=True,
        blank=True,)
    eve_region = models.ForeignKey(
        EveRegion,
        verbose_name=_("Region"),
        on_delete=models.CASCADE)
    min_volume = models.PositiveIntegerField(
        _("Minimum Volume"),
        null=True,
        blank=True,
        help_text=_("For buy orders, the minimum quantity that will be accepted in a matching sell order"))
    price = models.DecimalField(
        _("Price"),
        max_digits=15,
        decimal_places=2,
        help_text=_("Cost per unit for this order"))
    RANGE_CHOICES = [
        ('1', _('1')),
        ('2', _('2')),
        ('3', _('3')),
        ('4', _('4')),
        ('5', _('5')),
        ('10', _('10')),
        ('20', _('20')),
        ('30', _('30')),
        ('40', _('40')),
        ('region', _('Region')),
        ('solarsystem', _('Solar System')),
        ('station', _('Station'))]
    range = models.CharField(
        _("Order Range"),
        max_length=12,
        choices=RANGE_CHOICES,
        help_text=_("Valid order range, numbers are ranges in jumps"))
    volume_remain = models.PositiveIntegerField(
        _("Volume Remaining"),
        help_text=_("Quantity of items still required or offered"))
    is_corporation = models.BooleanField(
        _("Is Corporation"),
        default=False,
        help_text=_("Signifies whether the buy/sell order was placed on behalf of a corporation."))
    wallet_division = models.PositiveSmallIntegerField(
        _("Wallet Division"),
        null=True,
        blank=True,
        help_text=_("The corporation wallet division used for this order."))
    issued_by_character = models.ForeignKey(
        EveCharacter,
        verbose_name=_("Character"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True)
    issued_by_corporation = models.ForeignKey(
        EveCorporationInfo,
        verbose_name=_("Corporation"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,)
    STATE_CHOICES = [
        ('', ''),
        ('cancelled', _('Cancelled')),
        ('expired ', _('Expired'))]
    state = models.CharField(
        _("Order State"),
        max_length=10,
        choices=STATE_CHOICES,
        help_text=_("Current order state, Only valid for Authenticated order History. Will not update from Public Market Data."),
        null=True,
        blank=True,)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        """
        Meta definitions
        """
        default_permissions = ()

    @ property
    def expiry(self):
        return self.issued + datetime.timedelta(days=self.duration)


class Structure(models.Model):
    """An EVE Online Upwell Structure"""
    structure_id = models.PositiveBigIntegerField(
        _("Structure ID"),
        primary_key=True)
    name = models.CharField(
        _("Name"),
        max_length=60)
    owner_id = models.IntegerField(_("Owner Corporation ID"))
    solar_system = models.ForeignKey(
        EveSolarSystem,
        verbose_name=_("Solar System"),
        on_delete=models.CASCADE,
        related_name="+")
    eve_type = models.ForeignKey(
        EveType,
        verbose_name=_("Type"),
        on_delete=models.CASCADE,
        related_name="+",
        limit_choices_to={
            "eve_group__id__in": [15, 1657],
            "published": 1})
    pull_market = models.BooleanField(
        _("Pull Market Orders"),
        help_text=_("Useful to ignore specific structures for _reasons_"),
        default=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name


class Webhook(models.Model):
    name = models.CharField(
        _("Name"),
        max_length=100)
    url = models.URLField(
        _("URL"),
        max_length=200)
    enabled = models.BooleanField(default=True)

    def __str__(self):
        return self.name

    def send_embed(self, embed):
        webhook = SyncWebhook.from_url(self.url)
        webhook.send(embed=embed, username="Market Manager")


class Channel(models.Model):
    name = models.CharField(
        _("Name"),
        max_length=100)
    id = models.BigIntegerField(
        primary_key=True)

    def __str__(self):
        return self.name

    class Meta:
        """
        Meta definitions
        """
        verbose_name = _('Channel')
        verbose_name_plural = _('Channels')


class ManagedSupplyConfig(models.Model):
    managed_app = models.CharField(
        _("Managed By App"),
        max_length=50,
        default="",
        help_text=_("The App managing this SupplyConfig"),
        blank=True)
    managed_app_identifier = models.CharField(
        _("Managed App-Identifier"),
        max_length=50,
        default="",
        help_text=_("An identifier relevant to the App managing this SupplyConfig"),
        blank=True)
    managed_app_reason = models.CharField(
        _("Managed App Reason"),
        max_length=50,
        default="",
        help_text=_("User Facing reason for this Managed SupplyConfig"),
        blank=True)
    managed_quantity = models.IntegerField(
        _("quantity"),
        help_text=_("Apply a multiplier to the volumes that will be generated"),
        default=1,
        blank=False)
    # Duped from Supplyonfig
    managed_buy_order = models.BooleanField(_("Buy Order"))
    # Location Rules, Combined
    managed_structure = models.ManyToManyField(
        Structure,
        verbose_name=_("Structure"),
        blank=True)
    managed_solar_system = models.ManyToManyField(
        EveSolarSystem,
        verbose_name=_("Solar System"),
        blank=True)
    managed_region = models.ManyToManyField(
        EveRegion,
        verbose_name=_("Region"),
        blank=True)
    # Filter
    managed_structure_type = models.ManyToManyField(
        EveType,
        help_text=_("Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)"),
        verbose_name=_("Structure Type Filter"),
        blank=True)
    # Settings
    managed_jita_compare_percent = models.DecimalField(
        _("Jita Comparison %"),
        help_text=_("If set ignores Flat price value"),
        max_digits=4,
        decimal_places=0,
        default=Decimal(0),
        validators=[MinValueValidator(0), MaxValueValidator(1000)])
    # Destinations
    managed_webhooks = models.ManyToManyField(
        Webhook,
        verbose_name=_("Webhooks"),
        blank=True)
    managed_channels = models.ManyToManyField(
        Channel,
        verbose_name=_("Channels"),
        blank=True)
    managed_debug_webhooks = models.ManyToManyField(
        Webhook,
        verbose_name=_("Debug Webhook"),
        related_name="+",
        blank=True,
        help_text=_("Primarily for Testing/Debugging Purposes. This webhook will receive updates on Supplyonfigs that _dont_ notify. Because their Configs didnt Meet/Breach."))
    managed_debug_channels = models.ManyToManyField(
        Channel,
        verbose_name=_("Debug Channels"),
        related_name="+",
        blank=True,
        help_text=_("Primarily for Testing/Debugging Purposes. This webhook will receive updates on SupplyConfigs that _dont_ notify. Because their Configs didnt Meet/Breach."))

    def __str__(self):
        return self.managed_app_reason

    class Meta:
        """
        Meta definitions
        """
        verbose_name = _("Managed Supply Config")
        verbose_name_plural = _("Managed Supply Configs")


class SupplyConfig(models.Model):
    """Rules to Watch"""
    buy_order = models.BooleanField(_("Buy Order"))
    eve_type = models.ForeignKey(
        EveType,
        verbose_name=_("EVE Type"),
        on_delete=models.CASCADE,
        limit_choices_to={
            "eve_market_group__isnull": False,
            "published": 1})
    # Location Rules, Combined
    structure = models.ManyToManyField(
        Structure,
        verbose_name=_("Structure"),
        blank=True)
    solar_system = models.ManyToManyField(
        EveSolarSystem,
        verbose_name=_("Solar System"),
        blank=True)
    region = models.ManyToManyField(
        EveRegion,
        verbose_name=_("Region"),
        blank=True)
    # Filter
    structure_type = models.ManyToManyField(
        EveType,
        related_name="+",
        help_text=_("Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)"),
        verbose_name=_("Structure Type Filter"),
        blank=True)
    # Settings
    volume = models.IntegerField(
        _("Volume"),
        help_text=_("Set to Zero to check ANY/EVERY order against Price"),
        default=1,
        blank=False)
    price = models.DecimalField(
        _("Price"),
        help_text=_("Set to Zero to skip this filter"),
        default=Decimal(0),
        max_digits=15,
        decimal_places=2,
        blank=False)
    jita_compare_percent = models.DecimalField(
        _("Jita Comparison %"),
        help_text=_("If set ignores Flat price value"),
        max_digits=4,
        decimal_places=0,
        default=Decimal(0),
        validators=[MinValueValidator(0), MaxValueValidator(1000)])
    # Destinations
    webhooks = models.ManyToManyField(
        Webhook,
        verbose_name=_("Webhooks"),
        blank=True)
    channels = models.ManyToManyField(
        Channel,
        verbose_name=_("Channels"),
        blank=True)
    debug_webhooks = models.ManyToManyField(
        Webhook,
        verbose_name=_("Debug Webhook"),
        related_name="+",
        blank=True,
        help_text=_("Primarily for Testing/Debugging Purposes. This webhook will receive updates on SupplyConfigs that _dont_ notify. Because their Configs didnt Meet/Breach."))
    debug_channels = models.ManyToManyField(
        Channel,
        verbose_name=_("Debug Channels"),
        related_name="+",
        blank=True,
        help_text=_("Primarily for Testing/Debugging Purposes. This webhook will receive updates on SupplyConfigs that _dont_ notify. Because their Configs didnt Meet/Breach."))
    managed_supply_config = models.ForeignKey(
        ManagedSupplyConfig,
        verbose_name=_("Managed Watch Configs"),
        on_delete=models.CASCADE,
        blank=True,
        null=True)
    # Cached results from the last time this task was run
    last_run = models.DateTimeField(
        _("Last Task Runtime"),
        blank=True,
        null=True)
    last_result_volume = models.IntegerField(
        _("Last Result Volume"),
        help_text=_("The Volume returned the last time this WatchConfig was run"),
        blank=True,
        null=True)

    class Meta:
        """
        Meta definitions
        """
        verbose_name = _("Supply Config")
        verbose_name_plural = _("Supply Configs")


class PriceConfig(models.Model):
    """Rules to Watch"""
    buy_order = models.BooleanField(
        _("Buy Order"),
        default=False)
    scalp = models.BooleanField(
        _("Scalp Finder?"),
        help_text=_("True for Scalp, False for Bargain"),
        default=False)
    eve_type = models.ManyToManyField(
        EveType,
        verbose_name=_("EVE Type"),
        blank=True)
    eve_group = models.ManyToManyField(
        EveGroup,
        verbose_name=_("EVE Inventory Group"),
        blank=True)
    eve_market_group = models.ManyToManyField(
        EveMarketGroup,
        verbose_name=_("EVE Market Group"),
        blank=True)
    minimum = models.DecimalField(
        _("Minimum Price"),
        help_text=("Minimum isk to consider an order at, Default: 1M"),
        default=Decimal(1000000),
        max_digits=15,
        decimal_places=2,
        blank=False)
    # Location Rules, Combined
    structure = models.ManyToManyField(
        Structure,
        verbose_name=_("Structure"),
        blank=True)
    solar_system = models.ManyToManyField(
        EveSolarSystem,
        verbose_name=_("Solar System"),
        blank=True)
    region = models.ManyToManyField(
        EveRegion,
        verbose_name=_("Region"),
        blank=True)
    # Filter
    structure_type = models.ManyToManyField(
        EveType,
        related_name="+",
        help_text=_("Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)"),
        verbose_name=_("Structure Type Filter"),
        blank=True)
    # Settings
    jita_compare_percent = models.DecimalField(
        _("Jita Comparison %"),
        help_text=_("If set ignores Flat price value"),
        max_digits=4,
        decimal_places=0,
        default=Decimal(0),
        validators=[MinValueValidator(0), MaxValueValidator(1000)])
    price = models.DecimalField(
        _("Price"),
        help_text=_(
            "WARNING: Use Jita% when possible, this sets a flat price for ALL items included in this Config. Possibly useful for Caps/Soups"),
        default=Decimal(0),
        max_digits=15,
        decimal_places=2,
        blank=False)
    # Destinations
    webhooks = models.ManyToManyField(
        Webhook,
        verbose_name=_("Webhooks"),
        blank=True)
    channels = models.ManyToManyField(
        Channel,
        verbose_name=_("Channels"),
        blank=True)
    debug_webhooks = models.ManyToManyField(
        Webhook,
        verbose_name=_("Debug Webhook"),
        related_name="+",
        blank=True,
        help_text=_("Primarily for Testing/Debugging Purposes. This webhook will receive updates on PriceConfigs that _dont_ notify. Because their Configs didnt Meet/Breach."))
    debug_channels = models.ManyToManyField(
        Channel,
        verbose_name=_("Debug Channels"),
        related_name="+",
        blank=True,
        help_text=_("Primarily for Testing/Debugging Purposes. This webhook will receive updates on PriceConfigs that _dont_ notify. Because their Configs didnt Meet/Breach."))
    # Cached results from the last time this task was run
    last_run = models.DateTimeField(
        _("Last Task Runtime"),
        blank=True,
        null=True)

    class Meta:
        """
        Meta definitions
        """
        verbose_name = _("Price Config")
        verbose_name_plural = _("Price Configs")


class PublicConfig(SingletonModel):
    fetch_regions = models.ManyToManyField(
        EveRegion,
        verbose_name=_("Fetch Regions"),
    )

    def __str__(self):
        return "Public Market Configuration"

    class Meta:
        """
        Meta definitions
        """
        verbose_name = _("Public Market Configuration")


class StatisticsConfig(SingletonModel):
    calculate_regions = models.ManyToManyField(
        EveRegion,
        verbose_name=_("TypeStatistics Calculation Regions"),
    )

    def __str__(self):
        return "TypeStatistics Calculation Configuration"

    class Meta:
        """
        Meta definitions
        """
        verbose_name = _("TypeStatistics Calculation Configuration")


class PrivateConfig(models.Model):
    token = models.OneToOneField(
        Token,
        verbose_name=_("ESI Token"),
        on_delete=models.CASCADE,
    )
    valid_corporations = models.ManyToManyField(
        EveCorporationInfo,
        verbose_name=_("Valid Corporation Markets for this Token"),
        blank=True,
    )
    valid_structures = models.ManyToManyField(
        Structure,
        verbose_name=_("Valid Structure Markets for this Token"),
        blank=True,
    )
    failed = models.BooleanField(
        _("Disabled due to Failure, Check reason, adjust config and re-enable"))
    failure_reason = models.CharField(
        _("Failure Reason"),
        max_length=100,
        blank=True,
        default="Failure Reason")

    class Meta:
        """
        Meta definitions
        """
        verbose_name = _("Private Market Configuration")
        verbose_name_plural = _("Private Market Configurations")


class TypeStatistics(models.Model):
    eve_type = models.ForeignKey(
        EveType,
        verbose_name=_("Type"),
        on_delete=models.CASCADE
    )
    eve_region = models.ForeignKey(
        EveRegion,
        verbose_name=_("Region"),
        on_delete=models.CASCADE,
        blank=True,
        null=True
    )
    buy_fifth_percentile = models.DecimalField(
        _("Buy 5th Percentile"),
        max_digits=15,
        decimal_places=2
    )
    sell_fifth_percentile = models.DecimalField(
        _("Sell 5th Percentile"),
        max_digits=15,
        decimal_places=2
    )
    buy_weighted_average = models.DecimalField(
        _("Buy Weighted Average"),
        max_digits=15,
        decimal_places=2
    )
    sell_weighted_average = models.DecimalField(
        _("Sell Weighted Average"),
        max_digits=15,
        decimal_places=2
    )
    buy_median = models.DecimalField(
        _("Buy Median"),
        max_digits=15,
        decimal_places=2
    )
    sell_median = models.DecimalField(
        _("Sell Median"),
        max_digits=15,
        decimal_places=2
    )

    class Meta:
        """
        Meta definitions
        """
        default_permissions = ()
        verbose_name = _("Item Type Statistics")
        verbose_name_plural = _("Item Types Statistics")



# Generated by Django 4.0.3 on 2022-03-19 06:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0020_alter_order_duration_alter_order_price_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='typestatistics',
            name='new_eden',
            field=models.BooleanField(default=False, verbose_name='New Eden Wide'),
        ),
    ]



# Generated by Django 4.2.10 on 2024-05-11 11:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketmanager', '0051_alter_privateconfig_failure_reason'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='managedsupplyconfig',
            options={'verbose_name': 'Managed Supply Config', 'verbose_name_plural': 'Managed Supply Configs'},
        ),
        migrations.AlterModelOptions(
            name='priceconfig',
            options={'verbose_name': 'Price Config', 'verbose_name_plural': 'Price Configs'},
        ),
        migrations.AlterModelOptions(
            name='privateconfig',
            options={'verbose_name': 'Private Market Configuration', 'verbose_name_plural': 'Private Market Configurations'},
        ),
        migrations.AlterModelOptions(
            name='supplyconfig',
            options={'verbose_name': 'Supply Config', 'verbose_name_plural': 'Supply Configs'},
        ),
        migrations.AlterModelOptions(
            name='typestatistics',
            options={'default_permissions': (), 'verbose_name': 'Item Type Statistics', 'verbose_name_plural': 'Item Types Statistics'},
        ),
        migrations.AlterField(
            model_name='order',
            name='range',
            field=models.CharField(choices=[('1', '1'), ('2', '2'), ('3', '3'), ('4', '4'), ('5', '5'), ('10', '10'), ('20', '20'), ('30', '30'), ('40', '40'), ('region', 'Region'), ('solarsystem', 'Solar System'), ('station', 'Station')], help_text='Valid order range, numbers are ranges in jumps', max_length=12, verbose_name='Order Range'),
        ),
    ]
